import os
from simple_settings import settings
import subprocess

from . import utils, session
from .models.base import RunStatus


class BatchDfileRunner:
    """
    Batch-execute a list of pre-created d-files.

    Used to execute BMDS on a remote Windows server, if for example you're
    running a library on a non-Windows OS.
    """

    def __init__(self, inputs):
        self.tempfiles = utils.TempFileList()
        self.inputs = inputs

    def get_outfile(self, dfile, model_name):
        outfile = dfile.replace('.(d)', '.out')
        oo2 = outfile.replace('.out', '.002')

        # if not exponential, exit early
        if 'exponential' not in model_name.lower():
            return outfile

        # side-effect- cleanup other files created by exponential
        if os.path.exists(outfile):
            self.tempfiles.append(outfile)
        if os.path.exists(oo2):
            self.tempfiles.append(oo2)

        # get exponential model prefix
        prefix = model_name.split('-')[1]
        path, fn = os.path.split(outfile)
        outfile = os.path.join(path, prefix + fn)
        return outfile

    def execute_job(self, obj):
        """
        Execute the BMDS model and parse outputs if successful.
        """

        # get executable path
        exe = session.BMDS\
            .get_model(obj['bmds_version'], obj['model_name'])\
            .get_exe_path()

        # write dfile
        dfile = self.tempfiles.get_tempfile(prefix='bmds-dfile-', suffix='.(d)')
        with open(dfile, 'w') as f:
            f.write(obj['dfile'])

        outfile = self.get_outfile(dfile, obj['model_name'])
        oo2 = outfile.replace('.out', '.002')

        proc = subprocess.Popen(
            [exe, dfile],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)

        output = None
        stdout = ''
        stderr = ''

        try:
            stdout, stderr = proc.communicate(
                timeout=settings.BMDS_MODEL_TIMEOUT_SECONDS)

            if os.path.exists(outfile):
                with open(outfile, 'r') as f:
                    output = f.read()

            status = RunStatus.SUCCESS.value
            stdout = stdout.decode().strip()
            stderr = stderr.decode().strip()

        except subprocess.TimeoutExpired:
            proc.kill()
            status = RunStatus.FAILURE.value
            stdout, stderr = proc.communicate()

        finally:
            if os.path.exists(outfile):
                self.tempfiles.append(outfile)
            if os.path.exists(oo2):
                self.tempfiles.append(oo2)

        self.tempfiles.cleanup()

        return dict(
            status=status,
            output=output,
            stdout=stdout,
            stderr=stderr,
        )

    def execute(self):
        return list(map(self.execute_job, self.inputs))



from collections import defaultdict
import numpy as np
from simple_settings import settings
from scipy import stats

from . import plotting
from .anova import AnovaTests

__all__ = [
    'DichotomousDataset',
    'ContinuousDataset',
    'ContinuousIndividualDataset',
]


class Dataset(object):
    # Abstract parent-class for dataset-types.

    def _validate(self):
        raise NotImplemented('Abstract method; requires implementation')

    def as_dfile(self):
        raise NotImplemented('Abstract method; requires implementation')

    def to_dict(self):
        raise NotImplemented('Abstract method; requires implementation')

    def plot(self):
        raise NotImplemented('Abstract method; requires implementation')

    def drop_dose(self):
        raise NotImplemented('Abstract method; requires implementation')

    @property
    def num_dose_groups(self):
        return len(set(self.doses))

    def _get_dose_units_text(self):
        return ' ({})'.format(self.kwargs['dose_units']) \
            if 'dose_units' in self.kwargs \
            else ''

    def _get_response_units_text(self):
        return ' ({})'.format(self.kwargs['response_units']) \
            if 'response_units' in self.kwargs \
            else ''

    def _get_dataset_name(self):
        return self.kwargs.get('dataset_name', 'BMDS output results')


class DichotomousDataset(Dataset):
    """
    Dataset object for dichotomous datasets.

    A dichotomous dataset contains a list of 3 identically sized arrays of
    input values, for the dose, number of subjects, and incidences (subjects
    with a positive response).

    Example
    -------
    >>> dataset = bmds.DichotomousDataset(
            doses=[0, 1.96, 5.69, 29.75],
            ns=[75, 49, 50, 49],
            incidences=[5, 1, 3, 14]
        )
    """

    _BMDS_DATASET_TYPE = 1  # group data

    def __init__(self, doses, ns, incidences, **kwargs):
        self.doses = doses
        self.ns = ns
        self.incidences = incidences
        self.remainings = [n - p for n, p in zip(ns, incidences)]
        self.kwargs = kwargs
        self._sort_by_dose_group()
        self._validate()

    def _sort_by_dose_group(self):
        # use mergesort since it's a stable-sorting algorithm in numpy
        indexes = np.array(self.doses).argsort(kind='mergesort')
        for fld in ('doses', 'ns', 'incidences', 'remainings'):
            arr = getattr(self, fld)
            setattr(self, fld, np.array(arr)[indexes].tolist())
        self._validate()

    def _validate(self):
        length = len(self.doses)
        if not all(
                len(lst) == length for lst in
                [self.doses, self.ns, self.incidences]):
            raise ValueError('All input lists must be same length')

        if length != len(set(self.doses)):
            raise ValueError('Doses are not unique')

        if self.num_dose_groups < 3:
            raise ValueError('Must have 3 or more dose groups after dropping doses')

    def drop_dose(self):
        """
        Drop the maximum dose and related response values.
        """
        for fld in ('doses', 'ns', 'incidences', 'remainings'):
            arr = getattr(self, fld)[:-1]
            setattr(self, fld, arr)
        self._validate()

    def as_dfile(self):
        """
        Return the dataset representation in BMDS .(d) file.

        Example
        -------
        >>> print(dataset.as_dfile())
        Dose Incidence NEGATIVE_RESPONSE
        0.000000 5 70
        1.960000 1 48
        5.690000 3 47
        29.750000 14 35
        """
        rows = ['Dose Incidence NEGATIVE_RESPONSE']
        for i, v in enumerate(self.doses):
            if i >= self.num_dose_groups:
                continue
            rows.append('%f %d %d' % (
                self.doses[i], self.incidences[i], self.remainings[i]))
        return '\n'.join(rows)

    @property
    def dataset_length(self):
        """
        Return the length of the vector of doses-used.
        """
        return self.num_dose_groups

    def to_dict(self):
        """
        Returns a dictionary representation of the dataset.
        """
        d = dict(
            doses=self.doses,
            ns=self.ns,
            incidences=self.incidences,
        )
        d.update(self.kwargs)
        return d

    @staticmethod
    def _calculate_plotting(n, incidence):
        """
        Add confidence intervals to dichotomous datasets. From bmds231_manual.pdf, pg 124-5.

        LL = {(2np + z2 - 1) - z*sqrt[z2 - (2+1/n) + 4p(nq+1)]}/[2*(n+z2)]
        UL = {(2np + z2 + 1) + z*sqrt[z2 + (2-1/n) + 4p(nq-1)]}/[2*(n+z2)]

        - p = the observed proportion
        - n = the total number in the group in question
        - z = Z(1-alpha/2) is the inverse standard normal cumulative
              distribution function evaluated at 1-alpha/2
        - q = 1-p.

        The error bars shown in BMDS plots use alpha = 0.05 and so
        represent the 95% confidence intervals on the observed
        proportions (independent of model).
        """
        p = incidence / float(n)
        z = stats.norm.ppf(0.975)
        q = 1. - p
        ll = ((2 * n * p + 2 * z - 1) - z *
              np.sqrt(2 * z - (2 + 1 / n) + 4 * p * (n * q + 1))) / (2 * (n + 2 * z))
        ul = ((2 * n * p + 2 * z + 1) + z *
              np.sqrt(2 * z + (2 + 1 / n) + 4 * p * (n * q - 1))) / (2 * (n + 2 * z))
        return p, ll, ul

    def _set_plot_data(self):
        if hasattr(self, '_means'):
            return
        self._means, self._lls, self._uls = zip(*[
            self._calculate_plotting(i, j)
            for i, j in zip(self.ns, self.incidences)
        ])

    def plot(self):
        """
        Return a matplotlib figure of the dose-response dataset.

        Examples
        --------
        >>> fig = dataset.plot()
        >>> fig.show()
        >>> fig.clear()

        .. image:: ../tests/resources/test_ddataset_plot.png
           :align: center
           :alt: Example generated BMD plot

        Returns
        -------
        out : matplotlib.figure.Figure
            A matplotlib figure representation of the dataset.
        """
        self._set_plot_data()
        fig = plotting.create_empty_figure()
        ax = fig.gca()
        xlabel = self.kwargs.get('xlabel', 'Dose')
        ylabel = self.kwargs.get('ylabel', 'Fraction affected')
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.errorbar(
            self.doses, self._means, yerr=[self._lls, self._uls],
            label='Fraction affected ± 95% CI',
            **plotting.DATASET_POINT_FORMAT)
        ax.margins(plotting.PLOT_MARGINS)
        ax.set_title(self._get_dataset_name())
        ax.legend(**settings.LEGEND_OPTS)
        return fig


class ContinuousDataset(Dataset):
    """
    Dataset object for continuous datasets.

    A continuous dataset contains a list of 4 identically sized arrays of
    input values, for the dose, number of subjects, mean of response values for
    dose group, and standard-deviation of response for that dose group.

    Example
    -------
    >>> dataset = bmds.ContinuousDataset(
            doses=[0, 10, 50, 150, 400],
            ns=[25, 25, 24, 24, 24],
            means=[2.61, 2.81, 2.96, 4.66, 11.23],
            stdevs=[0.81, 1.19, 1.37, 1.72, 2.84]
        )
    """

    _BMDS_DATASET_TYPE = 1  # group data

    def __init__(self, doses, ns, means, stdevs, **kwargs):
        self.doses = doses
        self.ns = ns
        self.means = means
        self.stdevs = stdevs
        self.kwargs = kwargs
        self._sort_by_dose_group()
        self._validate()

    def _sort_by_dose_group(self):
        # use mergesort since it's a stable-sorting algorithm in numpy
        indexes = np.array(self.doses).argsort(kind='mergesort')
        for fld in ('doses', 'ns', 'means', 'stdevs'):
            arr = getattr(self, fld)
            setattr(self, fld, np.array(arr)[indexes].tolist())
        self._validate()

    def _validate(self):
        length = len(self.doses)
        if not all(
                len(lst) == length for lst in
                [self.doses, self.ns, self.means, self.stdevs]):
            raise ValueError('All input lists must be same length')

        if length != len(set(self.doses)):
            raise ValueError('Doses are not unique')

        if self.num_dose_groups < 3:
            raise ValueError('Must have 3 or more dose groups after dropping doses')

    @property
    def is_increasing(self):
        # increasing or decreasing with respect to control?
        change = 0.
        for i in range(1, len(self.means)):
            change += self.means[i] - self.means[0]
        return change >= 0

    def drop_dose(self):
        """
        Drop the maximum dose and related response values.
        """
        for fld in ('doses', 'ns', 'means', 'stdevs'):
            arr = getattr(self, fld)[:-1]
            setattr(self, fld, arr)
        self._validate()

    def as_dfile(self):
        """
        Return the dataset representation in BMDS .(d) file.
        """
        rows = ['Dose NumAnimals Response Stdev']
        for i, v in enumerate(self.doses):
            if i >= self.num_dose_groups:
                continue
            rows.append('%f %d %f %f' % (
                self.doses[i], self.ns[i], self.means[i], self.stdevs[i]))
        return '\n'.join(rows)

    @property
    def variances(self):
        if not hasattr(self, '_variances'):
            stds = np.array(self.stdevs)
            self._variances = np.power(stds, 2).tolist()
        return self._variances

    @property
    def anova(self):
        # Either be a tuple of 3 Test objects, or None if anova failed
        if not hasattr(self, '_anova'):
            try:
                num_params = 3  # assume linear model
                (A1, A2, A3, AR) = AnovaTests.compute_likelihoods(
                    self.num_dose_groups, self.ns, self.means, self.variances)
                tests = AnovaTests.get_anova_c3_tests(
                    num_params, self.num_dose_groups, A1, A2, A3, AR)
            except ValueError:
                tests = None
            self._anova = tests
        return self._anova

    @property
    def dataset_length(self):
        return self.num_dose_groups

    def get_anova_report(self):
        return AnovaTests.output_3tests(self.anova)

    def to_dict(self):
        """
        Return a dictionary representation of the dataset.
        """
        d = dict(
            doses=self.doses,
            ns=self.ns,
            means=self.means,
            stdevs=self.stdevs,
        )
        d.update(self.kwargs)
        return d

    @property
    def errorbars(self):
        # 95% confidence interval
        if not hasattr(self, '_errorbars'):
            self._errorbars = [
                stats.t.ppf(0.975, max(n - 1, 1)) * stdev / np.sqrt(float(n))
                for stdev, n in zip(self.stdevs, self.ns)
            ]
        return self._errorbars

    def plot(self):
        """
        Return a matplotlib figure of the dose-response dataset.

        Examples
        --------
        >>> fig = dataset.plot()
        >>> fig.show()

        .. image:: ../tests/resources/test_cdataset_plot.png
           :align: center
           :alt: Example generated BMD plot

        Returns
        -------
        out : matplotlib.figure.Figure
            A matplotlib figure representation of the dataset.
        """
        fig = plotting.create_empty_figure()
        ax = fig.gca()
        xlabel = self.kwargs.get('xlabel', 'Dose')
        ylabel = self.kwargs.get('ylabel', 'Response')
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.errorbar(
            self.doses, self.means, yerr=self.errorbars,
            label='Mean ± 95% CI',
            **plotting.DATASET_POINT_FORMAT)
        ax.margins(plotting.PLOT_MARGINS)
        ax.set_title(self._get_dataset_name())
        ax.legend(**settings.LEGEND_OPTS)
        return fig


class ContinuousIndividualDataset(ContinuousDataset):
    """
    Dataset object for continuous individual datasets.

    A continuous individual dataset contains a list of 2 identically sized
    arrays of input values, one for the dose and one for the response of an
    individual test-subject.

    Example
    -------
    >>> dataset = bmds.ContinuousIndividualDataset(
            doses=[
                0, 0, 0, 0, 0, 0, 0, 0,
                0.1, 0.1, 0.1, 0.1, 0.1, 0.1,
                1, 1, 1, 1, 1, 1,
                10, 10, 10, 10, 10, 10,
                100, 100, 100, 100, 100, 100,
                300, 300, 300, 300, 300, 300,
                500, 500, 500, 500, 500, 500,
            ],
            responses=[
                8.1079, 9.3063, 9.7431, 9.781, 10.052, 10.613, 10.751, 11.057,
                9.1556, 9.6821, 9.8256, 10.2095, 10.2222, 12.0382,
                9.5661, 9.7059, 9.9905, 10.2716, 10.471, 11.0602,
                8.8514, 10.0107, 10.0854, 10.5683, 11.1394, 11.4875,
                9.5427, 9.7211, 9.8267, 10.0231, 10.1833, 10.8685,
                10.368, 10.5176, 11.3168, 12.002, 12.1186, 12.6368,
                9.9572, 10.1347, 10.7743, 11.0571, 11.1564, 12.0368
            ]
        )
    """

    _BMDS_DATASET_TYPE = 0  # individual data

    def __init__(self, doses, responses, **kwargs):
        self.individual_doses = doses
        self.responses = responses
        self.kwargs = kwargs
        self._sort_by_dose_group()
        self.set_summary_data()
        self._validate()

    def _sort_by_dose_group(self):
        # use mergesort since it's a stable-sorting algorithm in numpy
        indexes = np.array(self.individual_doses).argsort(kind='mergesort')
        for fld in ('individual_doses', 'responses'):
            arr = getattr(self, fld)
            setattr(self, fld, np.array(arr)[indexes].tolist())

    def _validate(self):
        length = len(self.individual_doses)
        if not all(
                len(lst) == length for lst in
                [self.individual_doses, self.responses]):
            raise ValueError('All input lists must be same length')

        if self.num_dose_groups < 3:
            raise ValueError('Must have 3 or more doses after dropping doses')

    def set_summary_data(self):
        doses = list(set(self.individual_doses))
        doses.sort()

        dd = defaultdict(list)
        for d, r in zip(self.individual_doses, self.responses):
            dd[d].append(r)

        def _get_stats(lst):
            arr = np.array(lst, dtype=np.float64)
            return [arr.size, arr.mean(), arr.std()]

        vals = [_get_stats(dd[dose]) for dose in doses]
        self.ns, self.means, self.stdevs = zip(*vals)
        self.doses = doses

    def drop_dose(self):
        """
        Drop the maximum dose and related response values.
        """
        doses = np.array(self.individual_doses)
        responses = np.array(self.responses)
        mask = (doses != doses.max())
        self.individual_doses = doses[mask].tolist()
        self.responses = responses[mask].tolist()
        self.set_summary_data()
        self._validate()

    def as_dfile(self):
        """
        Return the dataset representation in BMDS .(d) file.
        """
        rows = ['Dose Response']
        for dose, response in zip(self.individual_doses, self.responses):
            dose_idx = self.doses.index(dose)
            if dose_idx >= self.num_dose_groups:
                continue
            rows.append('%f %f' % (dose, response))
        return '\n'.join(rows)

    def get_responses_by_dose(self):
        doses = np.array(self.individual_doses)
        resps = np.array(self.responses)
        return sorted([
            resps[doses == dose].tolist()
            for dose in self.doses
        ])

    @property
    def dataset_length(self):
        return len(self.individual_doses)

    def to_dict(self):
        """
        Return a dictionary representation of the dataset.
        """
        d = dict(
            individual_doses=self.individual_doses,
            responses=self.responses,
        )
        d.update(self.kwargs)
        return d

    def plot(self):
        """
        Return a matplotlib figure of the dose-response dataset.

        Examples
        --------
        >>> fig = dataset.plot()
        >>> fig.show()
        >>> fig.clear()

        .. image:: ../tests/resources/test_cidataset_plot.png
           :align: center
           :alt: Example generated BMD plot

        Returns
        -------
        out : matplotlib.figure.Figure
            A matplotlib figure representation of the dataset.
        """
        fig = plotting.create_empty_figure()
        ax = fig.gca()
        xlabel = self.kwargs.get('xlabel', 'Dose')
        ylabel = self.kwargs.get('ylabel', 'Response')
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.scatter(
            self.individual_doses, self.responses,
            label='Data',
            **plotting.DATASET_INDIVIDUAL_FORMAT)
        ax.margins(plotting.PLOT_MARGINS)
        ax.set_title(self._get_dataset_name())
        ax.legend(**settings.LEGEND_OPTS)
        return fig



import json
import os
import pandas as pd
from six import string_types

from .session import BMDS
from .reporter import Reporter


class SessionBatch(list):
    """
    Export utilities for exporting a collection of multiple BMD sessions.

    Example
    -------
    >>> datasets = [
            bmds.ContinuousDataset(
                doses=[0, 10, 50, 150, 400],
                ns=[111, 142, 143, 93, 42],
                means=[2.112, 2.095, 1.956, 1.587, 1.254],
                stdevs=[0.235, 0.209, 0.231, 0.263, 0.159]),
            bmds.ContinuousDataset(
                doses=[0, 10, 50, 150, 400],
                ns=[111, 142, 143, 93, 42],
                means=[2.112, 2.095, 1.956, 1.587, 1.254],
                stdevs=[0.235, 0.209, 0.231, 0.263, 0.159])
        ]
    >>> batch = bmds.SessionBatch()
        for dataset in datasets:
            session = bmds.BMDS.latest_version(
                bmds.constants.CONTINUOUS,
                dataset=dataset)
            session.add_default_models()
            session.execute()
            session.recommend()
            batch.append(session)
    >>> df = batch.to_df()
    >>> batch.to_csv('~/Desktop/outputs.csv')
    >>> batch.save_plots('~/Desktop', recommended_only=True)
    """

    def to_dicts(self):
        """
        Return a list of dictionaries of all model inputs and outputs.

        Parameters
        ----------
        filename : str or file
            Either the file name (string) or an open file (file-like object)
            where the data will be saved.

        Returns
        -------
        out : list
            List of output dictionaries.
        """
        return [session.to_dict(i) for i, session in enumerate(self)]

    def to_json(self, filename, indent=2):
        """
        Return a JSON string of all model inputs and outputs.

        Parameters
        ----------
        filename : str or file
            Either the file name (string) or an open file (file-like object)
            where the data will be saved.
        indent : int, optional
            Indentation level for JSON output.

        Returns
        -------
        out : str
            JSON formatted output string.

        """
        d = self.to_dicts()
        if hasattr(filename, 'write'):
            json.dump(d, filename, indent=indent)
        elif isinstance(filename, string_types):
            with open(os.path.expanduser(filename), 'w') as f:
                json.dump(d, f, indent=indent)
        else:
            raise ValueError('Unknown filename or file-object')

    def to_df(self, recommended_only=False, include_io=True):
        """
        Return a pandas DataFrame for each model and dataset.

        Parameters
        ----------
        recommended_only : bool, optional
            If True, only recommended models for each session are included. If
            no model is recommended, then a row with it's ID will be included,
            but all fields will be null.
        include_io :  bool, optional
            If True, then the input/output files from BMDS will also be
            included, specifically the (d) input file and the out file.

        Returns
        -------
        out : pandas.DataFrame
            Data frame containing models and outputs

        """
        od = BMDS._df_ordered_dict(include_io)
        [
            session._add_to_to_ordered_dict(od, i, recommended_only)
            for i, session in enumerate(self)
        ]
        return pd.DataFrame(od)

    def to_csv(self, filename, delimiter=',', recommended_only=False, include_io=True):
        """
        Return a CSV for each model and dataset.

        Parameters
        ----------
        filename : str or file
            Either the file name (string) or an open file (file-like object)
            where the data will be saved.
        delimiter : str, optional
            Delimiter used in CSV file between fields.
        recommended_only : bool, optional
            If True, only recommended models for each session are included. If
            no model is recommended, then a row with it's ID will be included,
            but all fields will be null.
        include_io :  bool, optional
            If True, then the input/output files from BMDS will also be
            included, specifically the (d) input file and the out file.

        Returns
        -------
        None

        """
        df = self.to_df(recommended_only, include_io)
        df.to_csv(filename, index=False, sep=delimiter)

    def to_excel(self, filename, recommended_only=False, include_io=True):
        """
        Return an Excel file for each model and dataset.

        Parameters
        ----------
        filename : str or ExcelWriter object
            Either the file name (string) or an ExcelWriter object.
        recommended_only : bool, optional
            If True, only recommended models for each session are included. If
            no model is recommended, then a row with it's ID will be included,
            but all fields will be null.
        include_io :  bool, optional
            If True, then the input/output files from BMDS will also be
            included, specifically the (d) input file and the out file.

        Returns
        -------
        None

        """
        df = self.to_df(recommended_only, include_io)
        if isinstance(filename, string_types):
            filename = os.path.expanduser(filename)
        df.to_excel(filename, index=False)

    def to_docx(self, filename=None,
                input_dataset=True, summary_table=True,
                recommendation_details=True, recommended_model=True,
                all_models=False):
        """
        Write batch sessions to a Word file.

        Parameters
        ----------
        filename : str or None
            If provided, the file is saved to this location, otherwise this
            method returns a docx.Document
        input_dataset : bool
            Include input dataset data table
        summary_table : bool
            Include model summary table
        recommendation_details : bool
            Include model recommendation details table
        recommended_model : bool
            Include the recommended model output and dose-response plot, if
            one exists
        all_models : bool
            Include all models output and dose-response plots

        Returns
        -------
        bmds.Reporter
            The bmds.Reporter object.

        """
        rep = Reporter()
        for model in self:
            rep.add_session(self, input_dataset, summary_table,
                            recommendation_details, recommended_model,
                            all_models)

        if filename:
            rep.save(filename)

        return rep

    def save_plots(self, directory, format='png', recommended_only=False):
        """
        Save images of dose-response curve-fits for each model.

        Parameters
        ----------
        directory : str
            Directory where the PNG files will be saved.
        format : str, optional
            Image output format. Valid options include: png, pdf, svg, ps, eps
        recommended_only : bool, optional
            If True, only recommended models for each session are included. If
            no model is recommended, then a row with it's ID will be included,
            but all fields will be null.

        Returns
        -------
        None

        """
        for i, session in enumerate(self):
            session.save_plots(
                directory,
                prefix=str(i),
                format=format,
                recommended_only=recommended_only)



class BMDSException(Exception):
    """An exception occurred in the BMDS module."""
    pass


class RemoteBMDSExcecutionException(BMDSException):
    """There was an error in executing BMDS on the BMDS server"""
    pass



"""
If we're not working on Windows, we cannot execute BMDS.

The BMDS source-code can be for other operating systems, but the compiled
binaries can yield different results from the same inputs. Thus, we treat the
Windows binaries as the "standard" and therefore we don't attempt to use
the alternative compilations.

Instead, we monkeypatch the executable steps in running a model and a session.
The dfiles are passed via HTTP to a remote Windows server which will execute
the BMDS and return the result .OUT files in a JSON format.

This requires an additional environment variable, BMDS_HOST, which is the host
path for remote execution: e.g., the string "http://12.13.145.167".
"""

from datetime import datetime
import json
import logging
import platform
import requests
import sys
from simple_settings import settings

from .session import BMDS
from .models.base import BMDModel, RunStatus
from .exceptions import RemoteBMDSExcecutionException


logger = logging.getLogger(__name__)
__all__ = []


def _get_payload(models):
    return dict(
        inputs=json.dumps([
            dict(
                bmds_version=model.bmds_version_dir,
                model_name=model.model_name,
                dfile=model.as_dfile(),
            ) for model in models
        ])
    )


if platform.system() != 'Windows':

    _request_session = None
    NO_HOST_WARNING = (
        'Using a non-Windows platform; BMDS cannot run natively in this OS.\n'
        'We can make a call to a remote server to execute.\n'
        'To execute BMDS, please specify the following environment variables:\n'
        '  - BMDS_HOST (e.g. http://bmds-server.com)\n'
        '  - BMDS_USERNAME (e.g. myusername)\n'
        '  - BMDS_PASSWORD (e.g. mysecret)\n'
    )

    def _get_requests_session():
        if settings.BMDS_HOST is None or \
           settings.BMDS_USERNAME is None or \
           settings.BMDS_PASSWORD is None:
                raise RemoteBMDSExcecutionException(NO_HOST_WARNING)

        global _request_session
        if _request_session is None:
            s = requests.Session()
            s.get('{}/admin/login/'.format(settings.BMDS_HOST))
            csrftoken = s.cookies['csrftoken']
            s.post('{}/admin/login/'.format(settings.BMDS_HOST), {
                'username': settings.BMDS_USERNAME,
                'password': settings.BMDS_PASSWORD,
                'csrfmiddlewaretoken': csrftoken,
            })

            # ensure authentication was successful
            if s.cookies.get('sessionid') is None:
                raise RemoteBMDSExcecutionException('Authentication failed')

            _request_session = s

        return _request_session

    def _set_results(model, results=None):
        if results is None:
            model._set_job_outputs(RunStatus.DID_NOT_RUN)
        else:
            status = results.pop('status')
            if status == RunStatus.SUCCESS:
                model._set_job_outputs(RunStatus.SUCCESS, **results)
            elif status == RunStatus.FAILURE:
                model._set_job_outputs(RunStatus.FAILURE)

    def execute_model(self):
        # execute single model
        self.execution_start = datetime.now()
        if self.can_be_executed:
            session = _get_requests_session()
            url = '{}/dfile/'.format(settings.BMDS_HOST)
            payload = _get_payload([self])
            logger.debug('Submitting payload: {}'.format(payload))
            resp = session.post(url, data=payload)
            results = resp.json()[0]
        else:
            results = None
        _set_results(self, results)

    def execute_session(self):
        # submit data
        start_time = datetime.now()
        executable_models = []
        for model in self.models:
            model.execution_start = start_time
            if model.can_be_executed:
                executable_models.append(model)
            else:
                _set_results(model)

        if len(executable_models) == 0:
            return

        session = _get_requests_session()
        url = '{}/dfile/'.format(settings.BMDS_HOST)
        payload = _get_payload(executable_models)
        logger.debug('Submitting payload: {}'.format(payload))
        resp = session.post(url, data=payload)

        # parse results for each model
        jsoned = resp.json()
        for model, results in zip(executable_models, jsoned):
            _set_results(model, results)

    # print startup error if host is None
    if settings.BMDS_HOST is None or \
       settings.BMDS_USERNAME is None or \
       settings.BMDS_PASSWORD is None:
            sys.stderr.write(NO_HOST_WARNING)

    # monkeypatch
    BMDS.execute = execute_session
    BMDModel.execute = execute_model



import re

from . import constants

EXPONENTIAL = 'E'
CTYPES = [constants.CONTINUOUS, constants.CONTINUOUS_INDIVIDUAL, EXPONENTIAL]


def try_float(v):
    try:
        return float(v)
    except ValueError:
        return v


class OutputParser(object):

    # regex for finding numeric values
    re_num = r'[/+/-]?[0-9]+[/.]*[0-9]*[Ee+-]*[0-9]*'

    # line-skips by dataset type
    NUM_LINE_SKIPS_PARAMS = {
        constants.CONTINUOUS: 1,
        constants.CONTINUOUS_INDIVIDUAL: 1,
        constants.DICHOTOMOUS: 1,
        constants.DICHOTOMOUS_CANCER: 1,
        EXPONENTIAL: 4,
    }

    NUM_LINE_SKIPS_FIT = {
        constants.CONTINUOUS: 3,
        constants.CONTINUOUS_INDIVIDUAL: 3,
        constants.DICHOTOMOUS: 2,
        constants.DICHOTOMOUS_CANCER: 2,
    }

    NUM_LINE_SKIPS_AIC = {
        constants.CONTINUOUS: 1,
        constants.CONTINUOUS_INDIVIDUAL: 1,
        EXPONENTIAL: 2,
    }

    def __init__(self, output_text, dtype, model_name):
        assert dtype in constants.DTYPES
        if 'exponential' in model_name.lower():
            dtype = EXPONENTIAL

        self.dtype = dtype
        self.output_text = output_text
        self.model_name = model_name
        self.parse_output_file()

    def parse_output_file(self):
        self.output = {}

        self._set_default_values()
        self._import_single_searches()
        self._import_warnings()
        self._import_line_by_line()
        self._calc_residual_of_interest()
        self._cleanup_nans()

    def _set_default_values(self):
        # set name
        self.output['model_name'] = self.model_name

        # initial setup
        vals = ('Chi2', 'df', 'p_value4', 'residual_of_interest')
        for val in vals:
            self.output[val] = -999

    def _import_line_by_line(self):
        if self.dtype in constants.DICHOTOMOUS_DTYPES:
            fit_tbl = (
                'fit_dose',
                'fit_est_prob',
                'fit_estimated',
                'fit_observed',
                'fit_size',
                'fit_residuals',
            )
        elif self.dtype in CTYPES:
            fit_tbl = (
                'fit_dose',
                'fit_size',
                'fit_observed',
                'fit_estimated',
                'fit_stdev',
                'fit_est_stdev',
                'fit_residuals'
            )

        # set empty tables
        for val in fit_tbl:
            self.output[val] = []

        # begin line-by-line
        outs = self.output_text.splitlines()
        if self.dtype in constants.DICHOTOMOUS_DTYPES:
            self._import_dich_vals()
            for i in range(len(outs)):
                if outs[i] == r'       Variable         Estimate        Std. Err.     Lower Conf. Limit   Upper Conf. Limit':  # noqa
                    self._lbl_parameter(outs, i)
                elif outs[i] == r'     Dose     Est._Prob.    Expected    Observed     Size       Residual':  # noqa
                    self._lbl_fit_cont_dich(outs, i, fit_tbl)

        elif self.dtype in constants.CONTINUOUS_DTYPES:
            for i in range(len(outs)):
                if outs[i] == r'       Variable         Estimate        Std. Err.     Lower Conf. Limit   Upper Conf. Limit':  # noqa
                    self._lbl_parameter(outs, i)
                elif outs[i] == r' Dose       N    Obs Mean     Est Mean   Obs Std Dev  Est Std Dev   Scaled Res.':  # noqa
                    self._lbl_fit_cont_dich(outs, i, fit_tbl)
                elif outs[i] == r'   Test    -2*log(Likelihood Ratio)  Test df        p-value    ':  # noqa
                    self._lbl_pvalue(outs, i)
                elif outs[i] == r"            Model      Log(likelihood)   # Param's      AIC":  # noqa
                    self._lbl_aic_cont_exp(outs, i)

        elif self.dtype == EXPONENTIAL:
            for i in range(len(outs)):
                if outs[i] == r'                     Parameter Estimates':
                    self._lbl_parameter(outs, i)
                elif outs[i] == r'     Dose      N         Obs Mean     Obs Std Dev':  # noqa
                    self._lbl_fit_exp(outs, i, 'observed_summary')
                elif outs[i] == r'     Dose      Obs Mean':  # noqa
                    self._lbl_fit_exp(outs, i, 'observed_individual')
                elif outs[i] == r'      Dose      Est Mean      Est Std     Scaled Residual':  # noqa
                    self._lbl_fit_exp(outs, i, 'estimated')
                elif outs[i] == r'     Test          -2*log(Likelihood Ratio)       D. F.         p-value':  # noqa
                    self._lbl_pvalue(outs, i)
                elif outs[i] == r"                     Model      Log(likelihood)      DF         AIC":  # noqa
                    self._lbl_aic_cont_exp(outs, i)

    def _cleanup_nans(self):
        # standardize possible errors
        fields = ('AIC', 'p_value1', 'p_value2', 'p_value3', 'p_value4')
        for field in fields:
            if field in self.output and self.output[field] in ['NA', 'N/A']:
                self.output[field] = -999

    def _calc_residual_of_interest(self):
        bmd = self.output['BMD']
        if bmd > 0 and len(self.output['fit_dose']) > 0:
            diff = abs(self.output['fit_dose'][0] - bmd)
            r = self.output['fit_residuals'][0]
            for j, val in enumerate(self.output['fit_dose']):
                if abs(val - bmd) < diff:
                    diff = abs(val - bmd)
                    r = self.output['fit_residuals'][j]
            self.output['residual_of_interest'] = r

    def _import_single_searches(self):
        """
        Look for simple one-line regex searches common across dataset types.

        If failed, then return -999.

        AIC is only handled in this method for dichotomous continuous and
        exponential and handled via separate methods.
        """
        searches = {
            # (?<!Setting ) is a special case for preventing
            # "Setting BMD = 100*(maximum dose)" matches
            'BMD': r'(?<!Setting )BMD = +(%s)' % self.re_num,
            'BMDL': r'BMDL = +(%s)' % self.re_num,
            'BMDU': r'BMDU = +(%s)' % self.re_num,
            'CSF': r'Cancer Slope Factor = +(%s)' % self.re_num,
            'AIC': r'AIC: +(%s)' % (self.re_num),
            'model_version': r'Version: ([\d\.]+);',
            'model_date': r'Date: ([\d/]+)',
        }
        for search in searches:
            m = re.search(searches[search], self.output_text)
            if m:
                try:
                    self.output[search] = float(m.group(1))
                except:
                    self.output[search] = m.group(1)
            else:
                self.output[search] = -999

    def _import_warnings(self):
        """
        Add custom warnings found in output files.

        Warnings in output files are searched for using this method; if a
        warning is found then it will be appended to the warnings list.
        """
        warnings = (
            r'Warning: BMDL computation is at best imprecise for these data',
            r'THE MODEL HAS PROBABLY NOT CONVERGED!!!',
            'THIS USUALLY MEANS THE MODEL HAS NOT CONVERGED!',
            r'BMR value is not in the range of the mean function',
            r'BMD = 100\*\(maximum dose\)',
            r'BMDL computation failed\.',
            'Warning:  optimum may not have been found.  Bad completion code in Optimization routine.',  # noqa
            'Warning: Likelihood for fitted model larger than the Likelihood for model A3.',  # noqa
        )
        self.output['warnings'] = []
        for warning in warnings:
            m = re.search(warning, self.output_text)
            if m:
                self.output['warnings'].append(m.group())

    def _import_dich_vals(self):
        """
        Import simple dichotomous values.

        Dichotomous values are found on one line, therefore one regex will
        return up to three possible matches for the Chi^2, degrees of freedom,
        and p-value.
        """
        m = re.search(r'Chi\^2 = ({0}|\w+) +d.f. = +({0}|\w+) +P-value = +({0}|\w+)'.format(self.re_num),  # noqa
                      self.output_text)
        cw = {1: "Chi2", 2: 'df', 3: 'p_value4'}
        for val in cw:
            try:
                self.output[cw[val]] = float(m.group(val))
            except:
                self.output[cw[val]] = -999

    def _lbl_parameter(self, outs, i):
        self.output['parameters'] = {}
        cw = {
            1: 'estimate',
            2: 'stdev',
            3: '95_low_limit',
            4: '95_high_limit',
        }
        i += self.NUM_LINE_SKIPS_PARAMS[self.dtype]
        while len(outs[i].split()) > 0:
            vals = outs[i].split()
            self.output['parameters'][vals[0]] = {}
            for j in range(1, len(vals)):
                try:
                    self.output['parameters'][vals[0]][cw[j]] = float(vals[j])
                except:
                    self.output['parameters'][vals[0]][cw[j]] = vals[j]
            i += 1

    def _lbl_fit_cont_dich(self, outs, i, fit_tbl):
        # Line-by-line: find "Goodness  of  Fit" table
        i += self.NUM_LINE_SKIPS_FIT[self.dtype]
        while i < len(outs) and len(outs[i]) > 1:
            vals = outs[i].split()
            for j in range(len(vals)):
                try:
                    self.output[fit_tbl[j]].append(float(vals[j]))
                except:
                    self.output[fit_tbl[j]].append(vals[j])
            i += 1

    def _lbl_fit_exp(self, outs, i, table_name):
        # Line-by-line: find "Goodness  of  Fit" table - exponential - observed
        i += 2   # next line and dotted lines
        if table_name == 'observed_summary':
            tbl_names = (1, 'fit_dose', 'fit_size', 'fit_observed', 'fit_stdev')
            rng = range(len(outs[i].split()))
        elif table_name == 'observed_individual':
            tbl_names = ('fit_dose', 'fit_observed')
            rng = range(1, len(outs[i].split()))
        elif table_name == 'estimated':
            tbl_names = ('fit_dose', 'fit_estimated', 'fit_est_stdev', 'fit_residuals')  # noqa
            rng = range(len(outs[i].split()))

        while i < len(outs) and len(outs[i]) > 0:
            vals = outs[i].split()
            # defensive code; parsing can fail if there's not enough text between
            # remove once BMDS 1.7 is released?
            if len(vals) == len(tbl_names):
                for j in rng:
                    self.output[tbl_names[j]].append(try_float(vals[j]))
            i += 1

    def _lbl_pvalue(self, outs, i):
        # Line-by-line: find p-values (continuous)
        i += 2   # next line and blank line
        while i < len(outs) and len(outs[i]) > 0:
            vals = outs[i].split()
            if vals[1] in ['5a', '6a', '7a']:  # fix exponentials
                pvalue = '4'
            else:
                pvalue = vals[1]

            if vals[4] == '<':
                self.output['p_value' + pvalue] = '<0.0001'
            else:
                try:
                    self.output['p_value' + pvalue] = float(vals[4])
                except:
                    if vals[4] == '<.0001':
                        self.output['p_value' + pvalue] = '<0.0001'
                    else:
                        self.output['p_value' + pvalue] = vals[4]

            if pvalue == '4':
                self.output['Chi2'] = try_float(vals[2])
                self.output['df'] = try_float(vals[3])
            i += 1

    def _lbl_aic_cont_exp(self, outs, i):
        # Line-by-line: find AIC (continuous)
        i += self.NUM_LINE_SKIPS_AIC[self.dtype]
        field = ('fitted', '2', '3', '4', '5')  # continuous is "fitted"
        while i < len(outs) and len(outs[i]) > 0:
            vals = outs[i].split()
            if vals[0] in field:
                try:
                    self.output['AIC'] = float(vals[3])
                except:
                    self.output['AIC'] = vals[3]
            i += 1



__version__ = '0.9.0'

import os  # noqa
os.environ.setdefault('SIMPLE_SETTINGS', 'bmds.settings')

from .session import * # noqa
from .datasets import *  # noqa
from .logic import *  # noqa
from .drunner import *  # noqa
from .batch import SessionBatch # noqa
from .reporter import Reporter, ReporterStyleGuide  # noqa
from .plotting import *  # noqa

from . import constants, models  # noqa

from .monkeypatch import * # noqa



from collections import namedtuple, OrderedDict
from io import BytesIO
import docx
from docx.shared import Inches
import numpy as np
import os
import re

from . import constants, datasets


ReporterStyleGuide = namedtuple('ReporterStyleGuide', [
    'table',
    'tbl_header',
    'tbl_body',
    'tbl_footnote',
    'outfile',
    'header_1',
    'header_2',
])


def default_float_formatter(value):
    if isinstance(value, str):
        return value
    elif np.isclose(value, int(value)):
        return str(int(value))
    elif abs(value) < 0.001:
        return '{:.1E}'.format(value)
    else:
        return '{:.3f}'.format(value).rstrip('0')


class TableFootnote(OrderedDict):
    def __init__(self):
        super().__init__()
        self.ascii_char = 96

    def add_footnote(self, p, text):
        if text not in self:
            self.ascii_char += 1
            self[text] = chr(self.ascii_char)
        self._add_footnote_character(p, self[text])

    def _add_footnote_character(self, p, symbol):
        run = p.add_run(symbol)
        run.font.superscript = True

    def add_footnote_text(self, doc, style):
        for text, char in self.items():
            p = doc.add_paragraph('', style=style)
            self._add_footnote_character(p, char)
            p.add_run(' {}'.format(text))


class Reporter:

    PORTRAIT_WIDTH = 6.5  # available width to use for tables

    def __init__(self, template=None, styles=None):
        """
        Create a new Microsoft Word document (.docx) reporter object.

        Parameters
        ----------
        template : str, docx.Document instance, or None.
            If str, the path the word template to be used.
            If a docx.Document object, this object is used.
            If None, the default template is used.

        styles : Instance of ReporterStyleGuide, or None.
            Determines which styles to apply for each component in document.
            If None, then the default ReporterStyleGuide is used. Note that if
            a custom template is specified, a ReporterStyleGuide should
            generally also be specified.

        Returns
        -------
        None.
        """

        if template is None:
            template = os.path.join(
                os.path.dirname(__file__),
                'templates/base.docx'
            )

        if styles is None:
            styles = ReporterStyleGuide(
                'bmdsTbl', 'bmdsTblHeader', 'bmdsTblBody',
                'bmdsTblFootnote', 'bmdsOutputFile', 'Heading 1', 'Heading 2'
            )

        self.styles = styles
        self.doc = docx.Document(template) \
            if isinstance(template, str) \
            else template

        # remove first paragraph if it's blank
        if len(self.doc.paragraphs) > 0 and \
                self.doc.paragraphs[0].text == '':
            self.doc._body._body.remove(self.doc.paragraphs[0]._p)

    def add_session(self, session,
                    input_dataset=True, summary_table=True,
                    recommendation_details=True, recommended_model=True,
                    all_models=False):
        """
        Add an existing session to a Word report.

        Parameters
        ----------
        session : bmds.Session
            BMDS session to be included in reporting
        input_dataset : bool
            Include input dataset data table
        summary_table : bool
            Include model summary table
        recommendation_details : bool
            Include model recommendation details table
        recommended_model : bool
            Include the recommended model output and dose-response plot, if
            one exists
        all_models : bool
            Include all models output and dose-response plots

        Returns
        -------
        None.
        """

        self.doc.add_paragraph(session.dataset._get_dataset_name(), self.styles.header_1)
        self.doc.add_paragraph('BMDS version: {}'.format(session.version_pretty))

        if input_dataset:
            self._add_dataset(session.dataset)
            self.doc.add_paragraph()

        if summary_table:
            self._add_session_summary_table(session)
            self.doc.add_paragraph()

        if recommendation_details:
            self._add_recommendation_details_table(session)
            self.doc.add_paragraph()

        if recommended_model and all_models:
            self._add_recommended_model(session)
            self._add_all_models(session, except_recommended=True)
            self.doc.add_paragraph()
        elif recommended_model:
            self._add_recommended_model(session)
            self.doc.add_paragraph()
        elif all_models:
            self._add_all_models(session, except_recommended=False)
            self.doc.add_paragraph()

        self.doc.add_page_break()

    def save(self, filename):
        """
        Save document to a file.

        Parameters
        ----------
        filename : str
            The output string filename

        """
        self.doc.save(os.path.expanduser(filename))

    def _set_col_width(self, column, size_in_inches):
        for cell in column.cells:
            cell.width = Inches(size_in_inches)

    def _add_dataset(self, dataset):

        self.doc.add_paragraph('Input dataset', self.styles.header_2)
        hdr = self.styles.tbl_header
        ff = default_float_formatter

        dose_units_text = dataset._get_dose_units_text()
        response_units_text = dataset._get_response_units_text()

        if isinstance(dataset, datasets.DichotomousDataset):

            tbl = self.doc.add_table(2, dataset.num_dose_groups + 1,
                                     style=self.styles.table)

            self._write_cell(tbl.cell(0, 0), 'Dose' + dose_units_text, style=hdr)
            self._write_cell(tbl.cell(1, 0), 'Affected / Total (%)' + response_units_text, style=hdr)
            for i, vals in enumerate(zip(dataset.doses,
                                         dataset.incidences,
                                         dataset.ns)):
                self._write_cell(tbl.cell(0, i + 1), vals[0])
                self._write_cell(
                    tbl.cell(1, i + 1),
                    '{}/{}\n({:.1%})'.format(vals[1], vals[2], vals[1] / float(vals[2]))
                )

            for i, col in enumerate(tbl.columns):
                w = 0.75 if i == 0 else \
                    (self.PORTRAIT_WIDTH - 0.75) / dataset.num_dose_groups
                self._set_col_width(col, w)

        elif isinstance(dataset, datasets.ContinuousIndividualDataset):

            tbl = self.doc.add_table(dataset.num_dose_groups + 1, 2,
                                     style=self.styles.table)

            self._write_cell(tbl.cell(0, 0), 'Dose' + dose_units_text, style=hdr)
            self._write_cell(tbl.cell(0, 1), 'Responses' + response_units_text, style=hdr)

            for i, vals in enumerate(zip(dataset.doses,
                                         dataset.get_responses_by_dose())):
                resps = ', '.join([ff(v) for v in vals[1]])
                self._write_cell(tbl.cell(i + 1, 0), vals[0])
                self._write_cell(tbl.cell(i + 1, 1), resps)

            self._set_col_width(tbl.columns[0], 1.0)
            self._set_col_width(tbl.columns[1], 5.5)

        elif isinstance(dataset, datasets.ContinuousDataset):

            tbl = self.doc.add_table(3, dataset.num_dose_groups + 1,
                                     style=self.styles.table)

            self._write_cell(tbl.cell(0, 0), 'Dose' + dose_units_text, style=hdr)
            self._write_cell(tbl.cell(1, 0), 'N', style=hdr)
            self._write_cell(tbl.cell(2, 0), 'Mean ± SD' + response_units_text, style=hdr)
            for i, vals in enumerate(zip(dataset.doses,
                                         dataset.ns,
                                         dataset.means,
                                         dataset.stdevs)):
                self._write_cell(tbl.cell(0, i + 1), vals[0])
                self._write_cell(tbl.cell(1, i + 1), vals[1])
                self._write_cell(tbl.cell(2, i + 1),
                                 '{} ± {}'.format(ff(vals[2]), ff(vals[3])))

            for i, col in enumerate(tbl.columns):
                w = 0.75 if i == 0 else \
                    (self.PORTRAIT_WIDTH - 0.75) / dataset.num_dose_groups
                self._set_col_width(col, w)

    def _write_cell(self, cell, value, style=None, float_formatter=None):

        if style is None:
            style = self.styles.tbl_body

        if isinstance(value, float):
            if float_formatter is None:
                float_formatter = default_float_formatter
                value = float_formatter(value)

        cell.paragraphs[0].text = str(value)
        cell.paragraphs[0].style = style

    _VARIANCE_FOOTNOTE_TEMPLATE = '{} case presented (BMDS Test 2 p-value = {}, BMDS Test 3 p-value = {}).'  # noqa

    def _get_variance_footnote(self, models):
        text = None
        for model in models:
            if model.has_successfully_executed:
                text = self._VARIANCE_FOOTNOTE_TEMPLATE.format(
                    model.get_variance_model_name(),
                    default_float_formatter(model.output['p_value2']),
                    default_float_formatter(model.output['p_value3']),
                )
                break
        if text is None:
            text = 'Model variance undetermined'
        return text

    def _get_summary_comments(self, session):
        if session.recommended_model is None:
            return 'No model was recommended as a best-fitting model.'
        else:
            return '{} recommended as best-fitting model on the basis of the lowest {}.'.format(
                session.recommended_model.name,
                session.recommended_model.recommended_variable
            )

    def _add_session_summary_table(self, session):
        self.doc.add_paragraph('Summary table', self.styles.header_2)
        hdr = self.styles.tbl_header
        model_groups = session._group_models()
        footnotes = TableFootnote()

        tbl = self.doc.add_table(len(model_groups) + 2, 6,
                                 style=self.styles.table)

        # write headers
        dose_units_text = session.dataset._get_dose_units_text()

        self._write_cell(tbl.cell(0, 0), 'Model', style=hdr)
        self._write_cell(tbl.cell(0, 1), 'Goodness of fit', style=hdr)
        self._write_cell(tbl.cell(1, 1), '', style=hdr)  # set style
        self._write_cell(tbl.cell(1, 2), 'AIC', style=hdr)
        self._write_cell(tbl.cell(0, 3), 'BMD' + dose_units_text, style=hdr)
        self._write_cell(tbl.cell(0, 4), 'BMDL' + dose_units_text, style=hdr)
        self._write_cell(tbl.cell(0, 5), 'Comments', style=hdr)

        p = tbl.cell(1, 1).paragraphs[0]
        p.add_run('p').italic = True
        p.add_run('-value')

        # add variance footnote to table header if appropriate
        if session.dtype in constants.CONTINUOUS_DTYPES:
            p = tbl.cell(0, 0).paragraphs[0]
            footnotes.add_footnote(p, self._get_variance_footnote(session.models))

        # merge header columns
        tbl.cell(0, 0).merge(tbl.cell(1, 0))
        tbl.cell(0, 1).merge(tbl.cell(0, 2))
        tbl.cell(0, 3).merge(tbl.cell(1, 3))
        tbl.cell(0, 4).merge(tbl.cell(1, 4))
        tbl.cell(0, 5).merge(tbl.cell(1, 5))

        # write body
        for i, model_group in enumerate(model_groups):
            idx = i + 2
            model = model_group[0]
            output = getattr(model, 'output', {})
            self._write_cell(tbl.cell(idx, 0), '')  # temp; set style
            self._write_cell(tbl.cell(idx, 1), output.get('p_value4', '-'))
            self._write_cell(tbl.cell(idx, 2), output.get('AIC', '-'))
            self._write_cell(tbl.cell(idx, 3), output.get('BMD', '-'))
            self._write_cell(tbl.cell(idx, 4), output.get('BMDL', '-'))

            self._write_model_name(tbl.cell(idx, 0).paragraphs[0],
                                   model_group, footnotes)

        # write comments
        self._write_cell(tbl.cell(2, 5), self._get_summary_comments(session))
        tbl.cell(2, 5).merge(tbl.cell(len(model_groups) + 1, 5))

        # set column width
        widths = [1.75, 0.8, 0.8, 0.7, 0.7, 1.75]
        for width, col in zip(widths, tbl.columns):
            self._set_col_width(col, width)

        # write footnote
        if len(footnotes) > 0:
            footnotes.add_footnote_text(self.doc, self.styles.tbl_footnote)

    def _add_recommendation_details_table(self, session):
        self.doc.add_paragraph('Model recommendation details',
                             self.styles.header_2)
        hdr = self.styles.tbl_header
        model_groups = session._group_models()
        footnotes = TableFootnote()

        tbl = self.doc.add_table(len(model_groups) + 1, 3,
                                 style=self.styles.table)

        # write headers
        self._write_cell(tbl.cell(0, 0), 'Model', style=hdr)
        self._write_cell(tbl.cell(0, 1), 'Bin', style=hdr)
        self._write_cell(tbl.cell(0, 2), 'Notes', style=hdr)

        def write_warnings(cell, txt, notes):
            p = cell.add_paragraph('')
            p.add_run(txt).bold = True

            if len(notes) == 0:
                cell.add_paragraph('  -')
            else:
                for note in notes:
                    cell.add_paragraph('• {}'.format(note))

        # write body
        for i, model_group in enumerate(model_groups):
            idx = i + 1
            model = model_group[0]
            bin = model.get_logic_bin_text().title()
            self._write_cell(tbl.cell(idx, 0), '')  # temp; set style
            self._write_model_name(tbl.cell(idx, 0).paragraphs[0], model_group,
                                   footnotes)
            self._write_cell(tbl.cell(idx, 1), bin)

            cell = tbl.cell(idx, 2)
            if not model.logic_notes[0] and \
                    not model.logic_notes[1] and \
                    not model.logic_notes[2]:
                cell.paragraphs[0].text = '-'
            else:
                cell._element.remove(cell.paragraphs[0]._p)
                if model.logic_notes[2]:
                    write_warnings(cell, 'Failures', model.logic_notes[2])
                if model.logic_notes[1]:
                    write_warnings(cell, 'Warnings', model.logic_notes[1])
                if model.logic_notes[0]:
                    write_warnings(cell, 'Cautions', model.logic_notes[0])

            for p in cell.paragraphs:
                p.style = self.styles.tbl_body

        # set column width
        widths = [1.75, 0.75, 4]
        for width, col in zip(widths, tbl.columns):
            self._set_col_width(col, width)

        # write footnote
        if len(footnotes) > 0:
            footnotes.add_footnote_text(self.doc, self.styles.tbl_footnote)

    def _write_model_name(self, p, model_group, footnotes):

        def pretty(name):
            name = name.replace('-', ' ')
            return re.sub(r'( \d+)', '\g<1>°', name)

        def collapse_names(models):
            names = [model.name for model in models]
            search_phrases = [
                'Polynomial-',
                'Multistage-Cancer-',
                'Multistage-'
            ]
            for phrase in search_phrases:
                if phrase in ''.join(names):
                    full_phrase = [
                        name for name in names
                        if phrase not in name
                    ]
                    matches = [
                        name for name in names
                        if phrase in name
                    ]
                    full_phrase.append(matches[0])
                    if len(matches) > 1:
                        remainders = ', '.join(matches[1:]).replace(phrase, '')
                        full_phrase.append(remainders)
                    names = full_phrase

            return ', '.join(names)

        p.add_run(pretty(model_group[0].name))

        if model_group[0].recommended:
            footnotes.add_footnote(p, 'Recommended model')

        if len(model_group) > 1:
            names = pretty(collapse_names(model_group[1:]))
            p.add_run(' (equivalent models include {})'.format(names))

    def _model_to_docx(self, model):
        if model.has_successfully_executed:
            # print figure
            fig = model.plot()
            with BytesIO() as f:
                fig.savefig(f)
                self.doc.add_picture(f, width=Inches(6))
            fig.clf()

            # print output file
            self.doc.add_paragraph(model.outfile, style=self.styles.outfile)
        else:
            self.doc.add_paragraph('No .OUT file was created.')

    def _add_recommended_model(self, session):
        self.doc.add_paragraph('Recommended model', self.styles.header_2)
        if hasattr(session, 'recommended_model') and \
                session.recommended_model is not None:
            self._model_to_docx(session.recommended_model)
        else:
            p = self.doc.add_paragraph()
            p.add_run('No model was recommended as a best-fitting model.')\
             .italic = True

    def _add_all_models(self, session, except_recommended=False):
        if except_recommended:
            self.doc.add_paragraph('All other models', self.styles.header_2)
        else:
            self.doc.add_paragraph('All model outputs', self.styles.header_2)

        for model in session.models:

            if model.recommended and except_recommended:
                continue

            self._model_to_docx(model)



# -*- coding: utf-8 -*-
import math

from .. import constants


class Rule(object):

    def __init__(self, failure_bin, **kwargs):
        self.failure_bin = failure_bin
        self.enabled = kwargs.get('enabled', True)
        self.threshold = kwargs.get('threshold', float('nan'))
        self.rule_name = kwargs.get('rule_name', self.default_rule_name)
        self.kwargs = kwargs

    def __unicode__(self):
        enabled = u'✓' if self.enabled else u'✕'
        threshold = '' if math.isnan(self.threshold) \
            else ', threshold={}'.format(self.threshold)
        return u'{0} {1} [bin={2}{3}]'.format(
            enabled, self.rule_name, self.binmoji, threshold)

    def check(self, dataset, output):
        if self.enabled:
            return self.apply_rule(dataset, output)
        else:
            return self.return_pass()

    @property
    def binmoji(self):
        return constants.BINMOJI[self.failure_bin]

    @property
    def bin_text(self):
        return constants.BIN_TEXT[self.failure_bin]

    def as_row(self):
        return [self.rule_name, self.enabled, self.bin_text, self.threshold]

    def return_pass(self):
        return constants.BIN_NO_CHANGE, None

    def apply_rule(self, dataset, output):
        # return tuple of (bin, notes) associated with rule or None
        raise NotImplementedError('Abstract method.')

    def get_failure_message(self, *args):
        raise NotImplementedError('Abstract method.')

    def _is_valid_number(self, val):
        # Ensure number is an int or float, not equal to special case -999.
        return val is not None and \
            val != -999 and \
            (isinstance(val, int) or isinstance(val, float))


class NumericValueExists(Rule):
    # Test succeeds if value is numeric and not -999
    field_name = None
    field_name_verbose = None

    def apply_rule(self, dataset, output):
        val = output.get(self.field_name)
        if self._is_valid_number(val):
            return self.return_pass()
        else:
            return self.failure_bin, self.get_failure_message()

    def get_failure_message(self):
        name = getattr(self, 'field_name_verbose')
        if name is None:
            name = self.field_name
        return '{} does not exist'.format(name)


class BmdExists(NumericValueExists):
    default_rule_name = 'BMD exists'
    field_name = 'BMD'


class BmdlExists(NumericValueExists):
    default_rule_name = 'BMDL exists'
    field_name = 'BMDL'


class BmduExists(NumericValueExists):
    default_rule_name = 'BMDU exists'
    field_name = 'BMDU'


class AicExists(NumericValueExists):
    default_rule_name = 'AIC exists'
    field_name = 'AIC'


class RoiExists(NumericValueExists):
    default_rule_name = 'Residual of interest exists'
    field_name = 'residual_of_interest'
    field_name_verbose = 'Residual of Interest'


class ShouldBeGreaterThan(Rule):
    # Test fails if value is less-than threshold.
    field_name = ''
    field_name_verbose = ''

    def apply_rule(self, dataset, output):
        val = output.get(self.field_name)
        threshold = self.threshold

        if not self._is_valid_number(val) or val >= threshold:
            return self.return_pass()
        else:
            return self.failure_bin, self.get_failure_message(val, threshold)

    def get_failure_message(self, val, threshold):
        name = self.field_name_verbose
        return '{} is less than threshold ({:.3} < {})'.format(name, float(val), threshold)


class VarianceFit(ShouldBeGreaterThan):
    default_rule_name = 'Variance fit'
    field_name = 'p_value3'
    field_name_verbose = 'Variance model fit p-value'


class GlobalFit(ShouldBeGreaterThan):
    default_rule_name = 'GGOF'
    field_name = 'p_value4'
    field_name_verbose = 'Goodness of fit p-value'


class ShouldBeLessThan(Rule):
    # Test fails if value is greater-than threshold.
    msg = ''  # w/ arguments for value and threshold

    def get_value(self, dataset, output):
        raise NotImplemented('Requires implementation')

    def apply_rule(self, dataset, output):
        val = self.get_value(dataset, output)
        threshold = self.threshold

        if not self._is_valid_number(val) or val <= threshold:
            return self.return_pass()
        else:
            return self.failure_bin, self.get_failure_message(val, threshold)

    def get_failure_message(self, val, threshold):
        name = self.field_name_verbose
        return '{} is greater than threshold ({:.3} > {})'.format(name, float(val), threshold)


class BmdBmdlRatio(ShouldBeLessThan):
    default_rule_name = 'BMD to BMDL ratio'
    field_name_verbose = 'BMD/BMDL ratio'

    def get_value(self, dataset, output):
        bmd = output.get('BMD')
        bmdl = output.get('BMDL')
        if self._is_valid_number(bmd) \
                and self._is_valid_number(bmdl) \
                and bmdl != 0:
            return bmd / bmdl


class RoiFit(ShouldBeLessThan):
    default_rule_name = 'Residual of interest'
    field_name_verbose = 'Residual of interest'

    def get_value(self, dataset, output):
        return output.get('residual_of_interest')


class HighBmd(ShouldBeLessThan):
    default_rule_name = 'High BMD'
    field_name_verbose = 'BMD/high dose ratio'

    def get_value(self, dataset, output):
        max_dose = max(dataset.doses)
        bmd = output.get('BMD')
        if self._is_valid_number(max_dose) \
                and self._is_valid_number(bmd)\
                and bmd != 0:
            return bmd / float(max_dose)


class HighBmdl(ShouldBeLessThan):
    default_rule_name = 'High BMDL'
    field_name_verbose = 'BMDL/high dose ratio'

    def get_value(self, dataset, output):
        max_dose = max(dataset.doses)
        bmdl = output.get('BMDL')
        if self._is_valid_number(max_dose) and \
                self._is_valid_number(bmdl) and \
                max_dose > 0:
            return bmdl / float(max_dose)


class LowBmd(ShouldBeLessThan):
    default_rule_name = 'Low BMD'
    field_name_verbose = 'BMD/minimum dose ratio'

    def get_value(self, dataset, output):
        min_dose = min([d for d in dataset.doses if d > 0])
        bmd = output.get('BMD')
        if self._is_valid_number(min_dose) and \
                self._is_valid_number(bmd) and \
                bmd > 0:
            return min_dose / float(bmd)


class LowBmdl(ShouldBeLessThan):
    default_rule_name = 'Low BMDL'
    field_name_verbose = 'BMDL/minimum dose ratio'

    def get_value(self, dataset, output):
        min_dose = min([d for d in dataset.doses if d > 0])
        bmdl = output.get('BMDL')
        if self._is_valid_number(min_dose) and \
                self._is_valid_number(bmdl) and \
                bmdl > 0:
            return min_dose / float(bmdl)


class ControlResidual(ShouldBeLessThan):
    default_rule_name = 'Control residual'
    field_name_verbose = 'Residual at lowest dose'

    def get_value(self, dataset, output):
        if output.get('fit_residuals') and len(output['fit_residuals']) > 0:
            try:
                return abs(output['fit_residuals'][0])
            except TypeError:
                return float('nan')


class ControlStdevResiduals(ShouldBeLessThan):
    default_rule_name = 'Control stdev'
    field_name_verbose = 'Ratio of modeled to actual stdev. at control'

    def get_value(self, dataset, output):
        if output.get('fit_est_stdev') and output.get('fit_stdev') and \
                len(output['fit_est_stdev']) > 0 and len(output['fit_stdev']) > 0:

            try:
                modeled = abs(output['fit_est_stdev'][0])
                actual = abs(output['fit_stdev'][0])
            except TypeError:
                return float('nan')

            if self._is_valid_number(modeled) and \
                    self._is_valid_number(actual) and \
                    modeled > 0 and actual > 0:
                return max(abs(modeled / actual), abs(actual / modeled))


class CorrectVarianceModel(Rule):
    # Check variance model (continuous datasets-only)
    default_rule_name = 'Variance type'

    def apply_rule(self, dataset, output):
        # 0 = non-homogeneous modeled variance => Var(i) = alpha*mean(i)^rho
        # 1 = constant variance => Var(i) = alpha*mean(i)
        if 'parameters' not in output:
            return self.return_pass()

        rho = output['parameters'].get('rho')
        constant_variance = 0 if rho else 1

        p_value2 = output.get('p_value2')
        if p_value2 == '<0.0001':
            p_value2 = 0.0001

        msg = None
        if self._is_valid_number(p_value2):
            if (constant_variance == 1 and p_value2 <= 0.1):
                msg = 'Incorrect variance model (p-value 2 = {}), constant variance selected'.format(p_value2)
            elif (constant_variance == 0 and p_value2 >= 0.1):
                msg = 'Incorrect variance model (p-value 2 = {}), modeled variance selected'.format(p_value2)
        else:
            msg = 'Correct variance model cannot be determined (p-value 2 = {})'.format(p_value2)

        if msg:
            return self.failure_bin, msg
        else:
            return self.return_pass()


class Warnings(Rule):
    # Test fails if any warnings exist.
    default_rule_name = 'Warnings'

    def get_failure_message(self, warnings):
        return u'Warning(s): {}'.format('; '.join(warnings))

    def apply_rule(self, dataset, output):
        warnings = output.get('warnings', [])
        if len(warnings) > 0:
            return self.failure_bin, self.get_failure_message(warnings)
        else:
            return self.return_pass()



from copy import deepcopy
import os
import numpy as np

from .base import BMDModel, DefaultParams
from .. import constants


class Continuous(BMDModel):
    dtype = 'C'  # for parsing output; therefore C is equivalent to CI
    possible_bmr = ('Abs. Dev.', 'Std. Dev.', 'Rel. Dev.', 'Point', 'Extra')

    def get_bmr_text(self):
        return '{} {}'.format(
            self.values['bmr'],
            constants.BMR_INVERTED_CROSSALK[self.dtype][self.values['bmr_type']],
        )

    def set_constant_variance_value(self):
        # 0 = non-homogeneous modeled variance => Var(i) = alpha*mean(i)^rho
        # 1 = constant variance => Var(i) = alpha*mean(i)
        return 0 if (self.dataset.anova is None or
                     self.dataset.anova[2].TEST <= 0.1) else 1

    def get_variance_model_name(self):
        return 'Modeled variance' \
            if self.values['constant_variance'] == 0 \
            else 'Constant variance'


# POLYNOMIAL
class Polynomial_216(Continuous):
    minimum_dose_groups = 2
    model_name = 'Polynomial'
    bmds_version_dir = 'BMDS231'
    exe = 'poly'
    exe_plot = '00poly'
    version = 2.16
    date = '05/26/2010'
    defaults = {
        'bmdl_curve_calculation': DefaultParams.bmdl_curve_calculation,
        'append_or_overwrite': DefaultParams.append_or_overwrite,
        'smooth_option': DefaultParams.smooth_option,
        'max_iterations': DefaultParams.max_iterations,
        'relative_fn_conv': DefaultParams.relative_fn_conv,
        'parameter_conv': DefaultParams.parameter_conv,
        'alpha': DefaultParams.param_generator('Alpha'),
        'rho': DefaultParams.param_generator('Rho'),
        'beta_0': DefaultParams.param_generator('Beta0'),
        'beta_1': DefaultParams.param_generator('Beta1'),
        'beta_2': DefaultParams.param_generator('Beta2'),
        'beta_3': DefaultParams.param_generator('Beta3'),
        'beta_4': DefaultParams.param_generator('Beta4'),
        'beta_5': DefaultParams.param_generator('Beta5'),
        'beta_7': DefaultParams.param_generator('Beta7'),
        'beta_6': DefaultParams.param_generator('Beta6'),
        'beta_8': DefaultParams.param_generator('Beta8'),
        'restrict_polynomial': DefaultParams.restrict(d=0, n='Restrict polynomial'),  # noqa
        'degree_poly': DefaultParams.degree_poly(),
        'bmd_calculation': DefaultParams.bmd_calculation,
        'dose_drop': DefaultParams.dose_drop,
        'bmr': DefaultParams.cont_bmr,
        'bmr_type': DefaultParams.cont_bmr_type,
        'confidence_level': DefaultParams.confidence_level,
        'constant_variance': DefaultParams.constant_variance,
    }

    @property
    def name(self):
        return u'{}-{}'.format(self.model_name, self._get_degrees())

    def set_restrict_polynomial_value(self):
        return 1 if self.dataset.is_increasing else -1

    def _get_degrees(self):
        degree = int(self.values['degree_poly'])
        if not 0 < degree <= 8:
            raise ValueError('Degree must be between 1 and 8, inclusive')
        return degree

    def as_dfile(self):
        self._set_values()
        degpoly = self._get_degrees()
        params = ['alpha', 'rho', 'beta_0']
        for i in range(1, degpoly + 1):
            params.append('beta_' + str(i))

        return '\n'.join([
            self._dfile_print_header_rows(),
            str(degpoly),
            '{} {} 0'.format(self.dataset._BMDS_DATASET_TYPE, self.dataset.dataset_length),
            self._dfile_print_options(
                'max_iterations', 'relative_fn_conv', 'parameter_conv',
                'bmdl_curve_calculation', 'restrict_polynomial',
                'bmd_calculation', 'append_or_overwrite', 'smooth_option'),
            self._dfile_print_options(
                'bmr_type', 'bmr', 'constant_variance', 'confidence_level'),
            self._dfile_print_parameters(*params),
            self.dataset.as_dfile(),
        ])

    def get_ys(self, xs):
        ys = np.zeros(xs.size)
        for i in range(self._get_degrees() + 1):
            param = self._get_param('beta_{}'.format(i))
            ys += np.power(xs, i) * param
        return ys


class Polynomial_217(Polynomial_216):
    bmds_version_dir = 'BMDS240'
    version = 2.17
    date = '01/28/2013'
    defaults = deepcopy(Polynomial_216.defaults)
    defaults['max_iterations']['d'] = 500


class Polynomial_220(Polynomial_217):
    bmds_version_dir = 'BMDS260'
    version = 2.20
    date = '10/22/2014'


class Polynomial_221(Polynomial_220):
    bmds_version_dir = 'BMDS270'
    version = 2.21
    date = '03/14/2017'


# LINEAR
class Linear_216(Polynomial_216):
    minimum_dose_groups = 2
    model_name = 'Linear'
    bmds_version_dir = 'BMDS231'
    exe = 'poly'
    exe_plot = '00poly'
    version = 2.16
    date = '05/26/2010'
    defaults = {
        'bmdl_curve_calculation': DefaultParams.bmdl_curve_calculation,
        'append_or_overwrite': DefaultParams.append_or_overwrite,
        'smooth_option': DefaultParams.smooth_option,
        'max_iterations': DefaultParams.max_iterations,
        'relative_fn_conv': DefaultParams.relative_fn_conv,
        'parameter_conv': DefaultParams.parameter_conv,
        'alpha': DefaultParams.param_generator('Alpha'),
        'rho': DefaultParams.param_generator('Rho'),
        'beta_0': DefaultParams.param_generator('Beta0'),
        'beta_1': DefaultParams.param_generator('Beta1'),
        'restrict_polynomial': DefaultParams.restrict(d=0, n='Restrict polynomial'),  # noqa
        'degree_poly': DefaultParams.degree_poly(d=1, showName=False),
        'bmd_calculation': DefaultParams.bmd_calculation,
        'dose_drop': DefaultParams.dose_drop,
        'bmr': DefaultParams.cont_bmr,
        'bmr_type': DefaultParams.cont_bmr_type,
        'confidence_level': DefaultParams.confidence_level,
        'constant_variance': DefaultParams.constant_variance,
    }

    @property
    def name(self):
        return self.model_name

    def as_dfile(self):
        self._set_values()
        return '\n'.join([
            self._dfile_print_header_rows(),
            '1',
            '{} {} 0'.format(self.dataset._BMDS_DATASET_TYPE, self.dataset.dataset_length),
            self._dfile_print_options(
                'max_iterations', 'relative_fn_conv', 'parameter_conv',
                'bmdl_curve_calculation', 'restrict_polynomial',
                'bmd_calculation', 'append_or_overwrite', 'smooth_option'),
            self._dfile_print_options(
                'bmr_type', 'bmr', 'constant_variance', 'confidence_level'),
            self._dfile_print_parameters(
                'alpha', 'rho', 'beta_0', 'beta_1'),
            self.dataset.as_dfile(),
        ])


class Linear_217(Linear_216):
    bmds_version_dir = 'BMDS240'
    version = 2.17
    date = '01/28/2013'
    defaults = deepcopy(Linear_216.defaults)
    defaults['max_iterations']['d'] = 500


class Linear_220(Linear_217):
    bmds_version_dir = 'BMDS260'
    version = 2.20
    date = '10/22/2014'


class Linear_221(Linear_220):
    bmds_version_dir = 'BMDS270'
    version = 2.21
    date = '03/14/2017'


# EXPONENTIAL M2
class Exponential(Continuous):

    def _get_model_name(self):
        return '{}_{}'.format(self.exe, self.output_prefix.lower())

    def get_outfile(self, dfile):
        # Exponential model output files play differently.
        #
        # Append the model-prefix to outfile.
        #
        # Note: this function has a side-effect of adding the blank out and 002
        # files which are created to be automatically cleaned-up.
        outfile = super().get_outfile(dfile)
        oo2 = outfile.replace('.out', '.002')
        if os.path.exists(outfile):
            self.tempfiles.append(outfile)
        if os.path.exists(oo2):
            self.tempfiles.append(oo2)
        path, fn = os.path.split(outfile)
        fn = self.output_prefix + fn
        return os.path.join(path, fn)

    def set_adverse_direction(self):
        return 1 if self.dataset.is_increasing else -1

    def as_dfile(self):
        self._set_values()
        params = self._dfile_print_parameters(
            'alpha', 'rho', 'a', 'b', 'c', 'd')
        return '\n'.join([
            self._dfile_print_header_rows(),
            self.exp_run_settings.format(
                self.dataset._BMDS_DATASET_TYPE,
                self.dataset.dataset_length,
                self.set_adverse_direction()),
            self._dfile_print_options(
                'max_iterations', 'relative_fn_conv', 'parameter_conv',
                'bmdl_curve_calculation', 'bmd_calculation',
                'append_or_overwrite', 'smooth_option'),
            self._dfile_print_options(
                'bmr_type', 'bmr', 'constant_variance', 'confidence_level'),
            '\n'.join([params] * 4),
            self.dataset.as_dfile(),
        ])


class Exponential_M2_17(Exponential):
    minimum_dose_groups = 2
    model_name = 'Exponential-M2'
    bmds_version_dir = 'BMDS231'
    exe = 'exponential'
    exe_plot = 'Expo_CPlot'
    exp_run_settings = '{} {} {} 1000 11 0 1'
    version = 1.7
    date = '12/10/2009'
    defaults = {
        'bmdl_curve_calculation': DefaultParams.bmdl_curve_calculation,
        'append_or_overwrite': DefaultParams.append_or_overwrite,
        'smooth_option': DefaultParams.smooth_option,
        'max_iterations': DefaultParams.max_iterations,
        'relative_fn_conv': DefaultParams.relative_fn_conv,
        'parameter_conv': DefaultParams.parameter_conv,
        'alpha': DefaultParams.param_generator('Alpha'),
        'rho': DefaultParams.param_generator('Rho'),
        'a': DefaultParams.param_generator('a'),
        'b': DefaultParams.param_generator('b'),
        'c': DefaultParams.param_generator('c'),
        'd': DefaultParams.param_generator('d'),
        'bmd_calculation': DefaultParams.bmd_calculation,
        'dose_drop': DefaultParams.dose_drop,
        'bmr': DefaultParams.cont_bmr,
        'bmr_type': DefaultParams.cont_bmr_type,
        'confidence_level': DefaultParams.confidence_level,
        'constant_variance': DefaultParams.constant_variance,
    }
    output_prefix = 'M2'

    def get_ys(self, xs):
        sign = 1. if self.dataset.is_increasing else -1.
        a = self._get_param('a')
        b = self._get_param('b')
        ys = a * np.exp(sign * b * xs)
        return ys


class Exponential_M2_19(Exponential_M2_17):
    bmds_version_dir = 'BMDS240'
    version = 1.9
    date = '01/29/2013'
    defaults = deepcopy(Exponential_M2_17.defaults)
    defaults['max_iterations']['d'] = 500


class Exponential_M2_110(Exponential_M2_19):
    bmds_version_dir = 'BMDS260'
    version = 1.10
    date = '01/12/2015'


class Exponential_M2_111(Exponential_M2_110):
    bmds_version_dir = 'BMDS270'
    version = 1.11
    date = '03/14/2017'


# EXPONENTIAL M3
class Exponential_M3_17(Exponential_M2_17):
    minimum_dose_groups = 3
    model_name = 'Exponential-M3'
    exp_run_settings = '{} {} {} 0100 22 0 1'
    output_prefix = 'M3'

    def get_ys(self, xs):
        sign = 1. if self.dataset.is_increasing else -1.
        a = self._get_param('a')
        b = self._get_param('b')
        d = self._get_param('d')
        ys = a * np.exp(sign * np.power(b * xs, d))
        return ys


class Exponential_M3_19(Exponential_M3_17):
    bmds_version_dir = 'BMDS240'
    version = 1.9
    ddate = '01/29/2013'
    defaults = deepcopy(Exponential_M3_17.defaults)
    defaults['max_iterations']['d'] = 500


class Exponential_M3_110(Exponential_M3_19):
    bmds_version_dir = 'BMDS260'
    version = 1.10
    date = '01/12/2015'


class Exponential_M3_111(Exponential_M3_110):
    bmds_version_dir = 'BMDS270'
    version = 1.11
    date = '03/14/2017'


# EXPONENTIAL M4
class Exponential_M4_17(Exponential_M2_17):
    minimum_dose_groups = 3
    model_name = 'Exponential-M4'
    exp_run_settings = '{} {} {} 0010 33 0 1'
    output_prefix = 'M4'

    def get_ys(self, xs):
        a = self._get_param('a')
        b = self._get_param('b')
        c = self._get_param('c')
        ys = a * (c - (c - 1.) * np.exp(-1. * b * xs))
        return ys


class Exponential_M4_19(Exponential_M4_17):
    bmds_version_dir = 'BMDS240'
    version = 1.9
    date = '01/29/2013'
    defaults = deepcopy(Exponential_M4_17.defaults)
    defaults['max_iterations']['d'] = 500


class Exponential_M4_110(Exponential_M4_19):
    bmds_version_dir = 'BMDS260'
    version = 1.10
    date = '01/12/2015'


class Exponential_M4_111(Exponential_M4_110):
    bmds_version_dir = 'BMDS270'
    version = 1.11
    date = '03/14/2017'


# EXPONENTIAL M5
class Exponential_M5_17(Exponential_M2_17):
    minimum_dose_groups = 4
    model_name = 'Exponential-M5'
    exp_run_settings = '{} {} {} 0001 44 0 1'
    output_prefix = 'M5'

    def get_ys(self, xs):
        a = self._get_param('a')
        b = self._get_param('b')
        c = self._get_param('c')
        d = self._get_param('d')
        ys = a * (c - (c - 1.) * np.exp(-1. * np.power(b * xs, d)))
        return ys


class Exponential_M5_19(Exponential_M5_17):
    bmds_version_dir = 'BMDS240'
    version = 1.9
    date = '01/29/2013'
    defaults = deepcopy(Exponential_M5_17.defaults)
    defaults['max_iterations']['d'] = 500


class Exponential_M5_110(Exponential_M5_19):
    bmds_version_dir = 'BMDS260'
    version = 1.10
    date = '01/12/2015'


class Exponential_M5_111(Exponential_M5_110):
    bmds_version_dir = 'BMDS270'
    version = 1.11
    date = '03/14/2017'


# POWER
class Power_216(Continuous):
    minimum_dose_groups = 3
    model_name = 'Power'
    bmds_version_dir = 'BMDS231'
    exe = 'power'
    exe_plot = '00power'
    version = 2.16
    date = '10/28/2009'
    defaults = {
        'bmdl_curve_calculation': DefaultParams.bmdl_curve_calculation,
        'append_or_overwrite': DefaultParams.append_or_overwrite,
        'smooth_option': DefaultParams.smooth_option,
        'log_transform': DefaultParams.log_transform(d=0),
        'max_iterations': DefaultParams.max_iterations,
        'relative_fn_conv': DefaultParams.relative_fn_conv,
        'parameter_conv': DefaultParams.parameter_conv,
        'alpha': DefaultParams.param_generator('Alpha'),
        'rho': DefaultParams.param_generator('Rho'),
        'control': DefaultParams.param_generator('Control'),
        'slope': DefaultParams.param_generator('Slope'),
        'power': DefaultParams.param_generator('Power'),
        'restrict_power': DefaultParams.restrict(d=1, n='Restrict power'),  # noqa
        'bmd_calculation': DefaultParams.bmd_calculation,
        'dose_drop': DefaultParams.dose_drop,
        'bmr': DefaultParams.cont_bmr,
        'bmr_type': DefaultParams.cont_bmr_type,
        'confidence_level': DefaultParams.confidence_level,
        'constant_variance': DefaultParams.constant_variance,
    }

    def as_dfile(self):
        self._set_values()
        return '\n'.join([
            self._dfile_print_header_rows(),
            '{} {} 0'.format(self.dataset._BMDS_DATASET_TYPE, self.dataset.dataset_length),
            self._dfile_print_options(
                'max_iterations', 'relative_fn_conv', 'parameter_conv',
                'bmdl_curve_calculation', 'restrict_power',
                'bmd_calculation', 'append_or_overwrite', 'smooth_option'),
            self._dfile_print_options(
                'bmr_type', 'bmr', 'constant_variance', 'confidence_level'),
            self._dfile_print_parameters(
                'alpha', 'rho', 'control', 'slope', 'power'),
            self.dataset.as_dfile(),
        ])

    def get_ys(self, xs):
        slope = self._get_param('slope')
        control = self._get_param('control')
        power = self._get_param('power')
        ys = control + slope * np.power(xs, power)
        return ys


class Power_217(Power_216):
    bmds_version_dir = 'BMDS240'
    version = 2.17
    date = '01/28/2013'
    defaults = deepcopy(Power_216.defaults)
    defaults['max_iterations']['d'] = 500


class Power_218(Power_217):
    bmds_version_dir = 'BMDS260'
    version = 2.18
    date = '05/19/2014'


class Power_219(Power_218):
    bmds_version_dir = 'BMDS270'
    version = 2.19
    date = '03/14/2017'


# HILL
class Hill_216(Continuous):
    minimum_dose_groups = 4
    model_name = 'Hill'
    bmds_version_dir = 'BMDS231'
    exe = 'hill'
    exe_plot = '00Hill'
    version = 2.16
    date = '04/06/2011'
    defaults = {
        'bmdl_curve_calculation': DefaultParams.bmdl_curve_calculation,
        'append_or_overwrite': DefaultParams.append_or_overwrite,
        'smooth_option': DefaultParams.smooth_option,
        'log_transform': DefaultParams.log_transform(d=0),
        'max_iterations': DefaultParams.max_iterations,
        'relative_fn_conv': DefaultParams.relative_fn_conv,
        'parameter_conv': DefaultParams.parameter_conv,
        'alpha': DefaultParams.param_generator('Alpha'),
        'rho': DefaultParams.param_generator('Rho'),
        'intercept': DefaultParams.param_generator('Intercept'),
        'v': DefaultParams.param_generator('V'),
        'n': DefaultParams.param_generator('N'),
        'k': DefaultParams.param_generator('K'),
        'restrict_n': DefaultParams.restrict(d=1, n='Restrict N>1'),
        'bmd_calculation': DefaultParams.bmd_calculation,
        'dose_drop': DefaultParams.dose_drop,
        'bmr': DefaultParams.cont_bmr,
        'bmr_type': DefaultParams.cont_bmr_type,
        'confidence_level': DefaultParams.confidence_level,
        'constant_variance': DefaultParams.constant_variance,
    }

    def as_dfile(self):
        self._set_values()
        return '\n'.join([
            self._dfile_print_header_rows(),
            '{} {} 0'.format(self.dataset._BMDS_DATASET_TYPE, self.dataset.dataset_length),
            self._dfile_print_options(
                'max_iterations', 'relative_fn_conv', 'parameter_conv',
                'bmdl_curve_calculation', 'restrict_n',
                'bmd_calculation', 'append_or_overwrite', 'smooth_option'),
            self._dfile_print_options(
                'bmr_type', 'bmr', 'constant_variance', 'confidence_level'),
            self._dfile_print_parameters(
                'alpha', 'rho', 'intercept', 'v', 'n', 'k'),
            self.dataset.as_dfile(),
        ])

    def get_ys(self, xs):
        intercept = self._get_param('intercept')
        v = self._get_param('v')
        n = self._get_param('n')
        k = self._get_param('k')
        ys = intercept + (v * np.power(xs, n)) / (np.power(k, n) + np.power(xs, n))
        return ys


class Hill_217(Hill_216):
    bmds_version_dir = 'BMDS240'
    version = 2.17
    date = '01/28/2013'
    defaults = deepcopy(Hill_216.defaults)
    defaults['max_iterations']['d'] = 500


class Hill_218(Hill_217):
    bmds_version_dir = 'BMDS270'
    version = 2.18
    date = '03/14/2017'



import os
import platform

# How long should the system wait for a BMDS model execution to complete
# before closing (in seconds)
BMDS_MODEL_TIMEOUT_SECONDS = 30

# Maximum polynomial used when setting default models; generally only used when
# the number of dose groups is large in a dataset. This is so that if you have
# 15 dose-groups you won't try to model using a 14 degree polynomial model.
# maximum_poly =  min(dataset dose groups - 1, MAXIMUM_POLYNOMIAL_ORDER)
MAXIMUM_POLYNOMIAL_ORDER = 8

# In model recommendation; if range of BMDLs is less than 3x different, the
# model with the smallest AIC will be recommended, otherwise the model with the
# smallest BMDL will be recommended.
SUFFICIENTLY_CLOSE_BMDL = 3

# Legend options applied to all plots which are generated in the software
LEGEND_OPTS = dict(
    loc='best',
    fontsize=8,
    fancybox=True,
    frameon=True,
)

# Only required if running BMDS on a non-Windows computer
if platform.system() != 'Windows':

    # The URL for the BMDS webserver
    BMDS_HOST = os.environ.get('BMDS_HOST')

    # The admin username for the BMDS webserver.
    BMDS_USERNAME = os.environ.get('BMDS_USERNAME')

    # The admin password for the BMDS webserver.
    BMDS_PASSWORD = os.environ.get('BMDS_PASSWORD')

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'default': {
            'format': '%(asctime)s %(levelname)s %(name)s %(message)s'
        },
        'simple': {
            'format': '%(levelname)s %(message)s'
        },
    },
    'handlers': {
        'logfile': {
            'level': 'DEBUG',
            'class': 'logging.handlers.RotatingFileHandler',
            'filename': 'bmds.log',
            'maxBytes': 50 * 1024 * 1024,
            'backupCount': 10,
            'formatter': 'default'
        },
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'simple',
        },
    },
    'loggers': {
        '': {
            'handlers': ['logfile'],
            'level': 'INFO',
        },
        'requests.packages.urllib3': {
            'handlers': ['logfile'],
            'level': 'INFO',
        }
    }
}

SIMPLE_SETTINGS = {
    'CONFIGURE_LOGGING': True,
}



import matplotlib as mpl
import matplotlib.pyplot as plt

__all__ = []

PLOT_FIGSIZE = (8, 5)
PLOT_MARGINS = 0.05
DATASET_POINT_FORMAT = dict(
    ms=7,
    fmt='o',
    c='k'
)
DATASET_INDIVIDUAL_FORMAT = dict(
    s=35,
    alpha=0.60,
    c='k'
)
LINE_FORMAT = dict(
    c='#6470C0',
    lw=3
)
BMD_LINE_FORMAT = dict(
    c='#BFC05D',
    lw=2
)
BMD_LABEL_FORMAT = dict(
    size=9
)
FAILURE_MESSAGE_FORMAT = dict(
    style='italic',
    weight='bold',
    bbox={'facecolor': 'red', 'alpha': 0.35, 'pad': 10},
    horizontalalignment='center',
    verticalalignment='center'
)


def create_empty_figure():
    plt.style.use('seaborn-darkgrid')
    mpl.rcParams.update({'font.size': 10})
    fig, ax = plt.subplots(figsize=PLOT_FIGSIZE)
    return fig



from .. import constants
from . import rules

import pandas as pd
from simple_settings import settings


class Recommender(object):
    # Recommendation logic for a specified data-type.

    def __init__(self, dtype, overrides=None):

        rule_args = dict()
        if dtype == constants.DICHOTOMOUS:
            pass
        elif dtype in constants.CONTINUOUS_DTYPES:
            rule_args['continuous'] = True
        elif dtype == constants.DICHOTOMOUS_CANCER:
            rule_args['cancer'] = True
        else:
            raise ValueError('Invalid data type')

        self.dtype = dtype
        self.rules = self._get_rule_defaults(**rule_args)
        self.SUFFICIENTLY_CLOSE_BMDL = settings.SUFFICIENTLY_CLOSE_BMDL

        if overrides:
            raise NotImplementedError('Overrides not implemented (yet).')

    @classmethod
    def _get_rule_defaults(cls, continuous=False, cancer=False):
        continous_only = True if continuous else False
        cancer_only = True if cancer else False
        ggof_threshold = 0.05 if cancer else 0.1
        return [
            rules.BmdExists(
                failure_bin=constants.BIN_FAILURE,
            ),
            rules.BmdlExists(
                failure_bin=constants.BIN_FAILURE,
            ),
            rules.BmduExists(
                failure_bin=constants.BIN_NO_CHANGE,
                enabled=cancer_only,
            ),
            rules.AicExists(
                failure_bin=constants.BIN_FAILURE,
            ),
            rules.RoiExists(
                failure_bin=constants.BIN_WARNING,
            ),
            rules.CorrectVarianceModel(
                failure_bin=constants.BIN_WARNING,
                threshold=0.1,
                enabled=continous_only,
            ),
            rules.VarianceFit(
                failure_bin=constants.BIN_WARNING,
                threshold=0.1,
                enabled=continous_only,
            ),
            rules.GlobalFit(
                failure_bin=constants.BIN_WARNING,
                threshold=ggof_threshold,
            ),
            rules.BmdBmdlRatio(
                failure_bin=constants.BIN_NO_CHANGE,
                threshold=5.,
                rule_name='BMD to BMDL ratio (warning)',
            ),
            rules.BmdBmdlRatio(
                failure_bin=constants.BIN_WARNING,
                threshold=20.,
            ),
            rules.RoiFit(
                failure_bin=constants.BIN_NO_CHANGE,
                threshold=2.,
            ),
            rules.Warnings(
                failure_bin=constants.BIN_NO_CHANGE,
            ),
            rules.HighBmd(
                failure_bin=constants.BIN_NO_CHANGE,
                threshold=1.,
            ),
            rules.HighBmdl(
                failure_bin=constants.BIN_WARNING,
                threshold=1.,
            ),
            rules.LowBmd(
                failure_bin=constants.BIN_NO_CHANGE,
                threshold=3.,
                rule_name='Low BMD (warning)',
            ),
            rules.LowBmdl(
                failure_bin=constants.BIN_NO_CHANGE,
                threshold=3.,
                rule_name='Low BMDL (warning)',
            ),
            rules.LowBmd(
                failure_bin=constants.BIN_WARNING,
                threshold=10.,
            ),
            rules.LowBmdl(
                failure_bin=constants.BIN_WARNING,
                threshold=10.,
            ),
            rules.ControlResidual(
                failure_bin=constants.BIN_WARNING,
                threshold=2.,
                enabled=continous_only,
            ),
            rules.ControlStdevResiduals(
                failure_bin=constants.BIN_WARNING,
                threshold=1.5,
                enabled=continous_only,
            ),
        ]

    def recommend(self, dataset, models):
        # apply rules to each model
        for model in models:

            # set global recommendations
            model.recommended = False
            model.recommended_variable = None

            # set no warnings by default (innocent until proven guilty)
            model.logic_notes = {
                constants.BIN_NO_CHANGE: [],
                constants.BIN_WARNING: [],
                constants.BIN_FAILURE: [],
            }

            # if no output is created, place model in failure bin
            if model.output_created:
                model.logic_bin = constants.BIN_NO_CHANGE
            else:
                model.logic_bin = constants.BIN_FAILURE
                continue

            # apply tests for each model
            for rule in self.rules:
                bin_, notes = rule.check(dataset, model.output)
                model.logic_bin = max(bin_, model.logic_bin)
                if notes:
                    model.logic_notes[bin_].append(notes)

        # get only models in highest bin-category
        model_subset = [
            model for model in models
            if model.logic_bin == constants.BIN_NO_CHANGE
        ]

        # exit early if there are no models left to recommend
        if len(model_subset) == 0:
            return

        # determine which approach to use for best-fitting model
        bmd_ratio = self._get_bmdl_ratio(model_subset)
        if bmd_ratio <= self.SUFFICIENTLY_CLOSE_BMDL:
            fld_name = 'AIC'
        else:
            fld_name = 'BMDL'

        # get and set recommended model
        model_subset = self._get_recommended_models(model_subset, fld_name)
        model = self._get_parsimonious_model(model_subset)
        model.recommended = True
        model.recommended_variable = fld_name
        return model

    def show_rules(self):
        return u'\n'.join([rule.__unicode__() for rule in self.rules])

    def rules_df(self):
        df = pd.DataFrame(
            data=[rule.as_row() for rule in self.rules],
            columns=['rule_name', 'enabled', 'failure_bin', 'threshold'])
        df = df[df.enabled == True]
        return df

    def _get_bmdl_ratio(self, models):
        """Return BMDL ratio in list of models."""
        bmdls = [
            model.output['BMDL']
            for model in models
            if model.output['BMDL'] > 0
        ]
        return max(bmdls) / min(bmdls)

    @staticmethod
    def _get_recommended_models(models, fld_name):
        """
        Returns a list of models which have the minimum target field value
        for a given field name (AIC or BMDL).
        """
        target_value = min([
            model.output[fld_name]
            for model in models
        ])
        return [
            model
            for model in models
            if model.output[fld_name] == target_value
        ]

    @staticmethod
    def _get_parsimonious_model(models):
        """
        Return the most parsimonious model of all available models. The most
        parsimonious model is defined as the model with the fewest number of
        parameters.
        """
        params = [
            len(model.output['parameters'])
            for model in models
        ]
        idx = params.index(min(params))
        return models[idx]



from math import log
from scipy import stats


class Test(object):

    def __init__(self):
        self.DF = -1
        self.CDF = -1
        self.SS = 0.0
        self.MSE = 0.0
        self.AIC = 0.0
        self.TEST = -1.0


class AnovaTests(object):

    @staticmethod
    def compute_likelihoods(n_obs, ns, ym, yd):
        Ntot = ns[0]
        Ntot = sum(ns)
        sigma2 = yd[0] * (ns[0] - 1)
        for i in range(1, n_obs):
            sigma2 += yd[i] * (ns[i] - 1)
        sigma2 = sigma2 / Ntot
        lkA1 = - Ntot * (1.0 + log(sigma2)) / 2.0
        lkA2 = - ns[0] * log(yd[0] * (ns[0] - 1) / ns[0]) / 2.0 - Ntot / 2.0
        for i in range(1, n_obs):
            lkA2 -= ns[i] * log(yd[i] * (ns[i] - 1) / ns[i]) / 2.0
        ybar = ym[0] * ns[0]
        for i in range(1, n_obs):
            ybar += ym[i] * ns[i]
        ybar = ybar / Ntot
        sigma2 = yd[0] * (ns[0] - 1) + ns[0] * (ym[0] - ybar) * (ym[0] - ybar)
        for i in range(1, n_obs):
            sigma2 += yd[i] * (ns[i] - 1) + \
                ns[i] * (ym[i] - ybar) * (ym[i] - ybar)
        sigma2 = sigma2 / Ntot
        lkR = -Ntot * (1.0 + log(sigma2)) / 2.0
        lkA3 = lkA1
        return lkA1, lkA2, lkA3, lkR

    @staticmethod
    def get_anova_c3_tests(nparm, n_obs, a1, a2, a3, ar):
        # Based on Hill model, modified from DTMS3ANOVAC.c

        # The xlk is not real
        xlk = (a1 + a2 + a3 + ar) / 4
        parm_known = 1
        anovaList = [Test() for i in range(5)]

        # Compute DF and assign LLK for each test
        anovaList[0].DF = n_obs + 1
        anovaList[0].SS = a1

        anovaList[1].DF = 2 * n_obs
        anovaList[1].SS = a2

        anovaList[2].DF = n_obs + 2 - parm_known
        anovaList[2].SS = a3

        anovaList[3].DF = 2
        anovaList[3].SS = ar

        anovaList[4].DF = nparm - 2
        anovaList[4].SS = xlk

        # Compute likelihood ratio MSE and CDF
        anovaList[0].MSE = 2 * (a2 - a1)
        anovaList[0].CDF = anovaList[1].DF - anovaList[0].DF

        anovaList[1].MSE = 2 * (a2 - a3)
        anovaList[1].CDF = anovaList[1].DF - anovaList[2].DF

        anovaList[2].MSE = 2 * (a3 - xlk)
        anovaList[2].CDF = anovaList[2].DF - anovaList[4].DF

        anovaList[3].MSE = 2 * (a2 - ar)
        anovaList[3].CDF = anovaList[1].DF - anovaList[3].DF

        for anova in anovaList:
            anova.AIC = -2 * (anova.SS - anova.DF)
            if (anova.MSE >= 0.0 and anova.CDF > 0):
                anova.TEST = 1 - stats.chi2.cdf(anova.MSE, anova.CDF)

        # Only return test 1, test 2 and test 3 in the order
        return (anovaList[3], anovaList[0], anovaList[1])

    @staticmethod
    def output_3tests(tests):
        if tests is None:
            return 'ANOVA cannot be calculated for this dataset.'

        outputs = [
            '                     Tests of Interest    ',
            '   Test    -2*log(Likelihood Ratio)  Test df        p-value    ',
        ]
        for i, test in enumerate(tests):
            outputs.append('   Test %d %20.6g %10d %16.4g' % (
                i + 1, test.MSE, test.CDF, test.TEST))
        return '\n'.join(outputs)



from copy import deepcopy
import numpy as np
from scipy.stats import norm, gamma

from .base import BMDModel, DefaultParams
from .. import constants


class Dichotomous(BMDModel):
    dtype = constants.DICHOTOMOUS
    possible_bmr = ('Extra', 'Added')

    def get_bmr_text(self):
        return '{:.0%} {} risk'.format(
            self.values['bmr'],
            constants.BMR_INVERTED_CROSSALK[self.dtype][self.values['bmr_type']],
        )


class DichotomousCancer(Dichotomous):
    dtype = constants.DICHOTOMOUS_CANCER
    possible_bmr = ('Extra', 'Added')


# MULTISTAGE
class Multistage_32(Dichotomous):
    minimum_dose_groups = 2
    model_name = 'Multistage'
    bmds_version_dir = 'BMDS231'
    exe = 'multistage'
    exe_plot = '10multista'
    version = 3.2
    date = '05/26/2010'
    defaults = {
        'bmdl_curve_calculation': DefaultParams.bmdl_curve_calculation,
        'append_or_overwrite': DefaultParams.append_or_overwrite,
        'smooth_option': DefaultParams.smooth_option,
        'max_iterations': DefaultParams.max_iterations,
        'relative_fn_conv': DefaultParams.relative_fn_conv,
        'parameter_conv': DefaultParams.parameter_conv,
        'background': DefaultParams.param_generator(name='Background'),
        'beta1': DefaultParams.param_generator(name='Beta1'),
        'beta2': DefaultParams.param_generator(name='Beta2'),
        'beta3': DefaultParams.param_generator(name='Beta3'),
        'beta4': DefaultParams.param_generator(name='Beta4'),
        'beta5': DefaultParams.param_generator(name='Beta5'),
        'beta6': DefaultParams.param_generator(name='Beta6'),
        'beta7': DefaultParams.param_generator(name='Beta7'),
        'beta8': DefaultParams.param_generator(name='Beta8'),
        'restrict_beta': DefaultParams.restrict(d=1, n='Restrict beta'),
        'degree_poly': DefaultParams.degree_poly(),
        'bmd_calculation': DefaultParams.bmd_calculation,
        'dose_drop': DefaultParams.dose_drop,
        'bmr': DefaultParams.dich_bmr,
        'bmr_type': DefaultParams.dich_bmr_type,
        'confidence_level': DefaultParams.confidence_level,
    }

    @property
    def name(self):
        return u'{}-{}'.format(self.model_name, self._get_degrees())

    def _get_degrees(self):
        degree = int(self.values['degree_poly'])
        if not 0 < degree <= 8:
            raise ValueError('Degree must be between 1 and 8, inclusive')
        return degree

    def as_dfile(self):
        self._set_values()
        degree_poly = self._get_degrees()

        params = ['beta{}'.format(i) for i in range(1, degree_poly + 1)]
        params.insert(0, 'background')

        return '\n'.join([
            self._dfile_print_header_rows(),
            '{} {}'.format(self.dataset.dataset_length, degree_poly),
            self._dfile_print_options(
                'max_iterations', 'relative_fn_conv', 'parameter_conv',
                'bmdl_curve_calculation', 'restrict_beta',
                'bmd_calculation', 'append_or_overwrite', 'smooth_option'),
            self._dfile_print_options(
                'bmr', 'bmr_type', 'confidence_level'),
            self._dfile_print_parameters(*params),
            self.dataset.as_dfile(),
        ])

    def get_ys(self, xs):
        background = self._get_param('Background')
        ys = np.zeros(xs.size)
        for i in range(1, self._get_degrees() + 1):
            param = self._get_param('Beta({})'.format(i))
            ys += param * np.power(xs, i)
        ys = background + (1.0 - background) * (1.0 - np.exp(-1.0 * ys))
        return ys


class Multistage_33(Multistage_32):
    bmds_version_dir = 'BMDS240'
    version = 3.3
    date = '02/28/2013'
    defaults = deepcopy(Multistage_32.defaults)
    defaults['max_iterations']['d'] = 500


class Multistage_34(Multistage_33):
    bmds_version_dir = 'BMDS260'
    version = 3.4
    date = '05/02/2014'


# MULTISTAGE CANCER
class MultistageCancer_19(DichotomousCancer):
    minimum_dose_groups = 2
    model_name = 'Multistage-Cancer'
    bmds_version_dir = 'BMDS231'
    exe = 'cancer'
    exe_plot = '10cancer'
    version = 1.9
    date = '05/26/2010'
    defaults = {
        'bmdl_curve_calculation': DefaultParams.bmdl_curve_calculation,
        'append_or_overwrite': DefaultParams.append_or_overwrite,
        'smooth_option': DefaultParams.smooth_option,
        'max_iterations': DefaultParams.max_iterations,
        'relative_fn_conv': DefaultParams.relative_fn_conv,
        'parameter_conv': DefaultParams.parameter_conv,
        'background': DefaultParams.param_generator(name='Background'),
        'beta1': DefaultParams.param_generator(name='Beta1'),
        'beta2': DefaultParams.param_generator(name='Beta2'),
        'beta3': DefaultParams.param_generator(name='Beta3'),
        'beta4': DefaultParams.param_generator(name='Beta4'),
        'beta5': DefaultParams.param_generator(name='Beta5'),
        'beta6': DefaultParams.param_generator(name='Beta6'),
        'beta7': DefaultParams.param_generator(name='Beta7'),
        'beta8': DefaultParams.param_generator(name='Beta8'),
        'restrict_beta': DefaultParams.restrict(d=1, n='Restrict beta'),
        'degree_poly': DefaultParams.degree_poly(),
        'bmd_calculation': DefaultParams.bmd_calculation,
        'dose_drop': DefaultParams.dose_drop,
        'bmr': DefaultParams.dich_bmr,
        'bmr_type': DefaultParams.dich_bmr_type,
        'confidence_level': DefaultParams.confidence_level,
    }

    @property
    def name(self):
        return u'{}-{}'.format(self.model_name, self._get_degrees())

    def _get_degrees(self):
        degree = int(self.values['degree_poly'])
        if not 0 < degree <= 8:
            raise ValueError('Degree must be between 1 and 8, inclusive')
        return degree

    def as_dfile(self):
        self._set_values()
        degree_poly = self._get_degrees()

        params = ['beta{}'.format(i) for i in range(1, degree_poly + 1)]
        params.insert(0, 'background')

        return '\n'.join([
            self._dfile_print_header_rows(),
            '{} {}'.format(self.dataset.dataset_length, degree_poly),
            self._dfile_print_options(
                'max_iterations', 'relative_fn_conv', 'parameter_conv',
                'bmdl_curve_calculation', 'restrict_beta',
                'bmd_calculation', 'append_or_overwrite', 'smooth_option'),
            self._dfile_print_options(
                'bmr', 'bmr_type', 'confidence_level'),
            self._dfile_print_parameters(*params),
            self.dataset.as_dfile(),
        ])

    def get_ys(self, xs):
        background = self._get_param('Background')
        ys = np.zeros(xs.size)
        for i in range(1, self._get_degrees() + 1):
            param = self._get_param('Beta({})'.format(i))
            ys += param * np.power(xs, i)
        ys = background + (1.0 - background) * (1.0 - np.exp(-1.0 * ys))
        return ys


class MultistageCancer_110(MultistageCancer_19):
    bmds_version_dir = 'BMDS240'
    version = 1.10
    date = '02/28/2013'
    defaults = deepcopy(MultistageCancer_19.defaults)
    defaults['max_iterations']['d'] = 500


class MultistageCancer_34(MultistageCancer_19):
    bmds_version_dir = 'BMDS270'
    version = 3.4
    date = '05/02/2014'


# WEIBULL
class Weibull_215(Dichotomous):
    minimum_dose_groups = 3
    model_name = 'Weibull'
    bmds_version_dir = 'BMDS231'
    exe = 'weibull'
    exe_plot = '10weibull'
    version = 2.15
    date = '10/28/2009'
    defaults = {
        'bmdl_curve_calculation': DefaultParams.bmdl_curve_calculation,
        'append_or_overwrite': DefaultParams.append_or_overwrite,
        'smooth_option': DefaultParams.smooth_option,
        'log_transform': DefaultParams.log_transform(d=0),
        'max_iterations': DefaultParams.max_iterations,
        'relative_fn_conv': DefaultParams.relative_fn_conv,
        'parameter_conv': DefaultParams.parameter_conv,
        'background': DefaultParams.param_generator(),
        'slope': DefaultParams.param_generator(name='Slope'),
        'power': DefaultParams.param_generator(name='Power'),
        'restrict_power': DefaultParams.restrict(d=1, n='Restrict power'),
        'bmd_calculation': DefaultParams.bmd_calculation,
        'dose_drop': DefaultParams.dose_drop,
        'bmr': DefaultParams.dich_bmr,
        'bmr_type': DefaultParams.dich_bmr_type,
        'confidence_level': DefaultParams.confidence_level,
    }

    def as_dfile(self):
        self._set_values()
        return '\n'.join([
            self._dfile_print_header_rows(),
            str(self.dataset.dataset_length),
            self._dfile_print_options(
                'max_iterations', 'relative_fn_conv', 'parameter_conv',
                'bmdl_curve_calculation', 'restrict_power', 'bmd_calculation',
                'append_or_overwrite', 'smooth_option'),
            self._dfile_print_options(
                'bmr', 'bmr_type', 'confidence_level'),
            self._dfile_print_parameters(
                'background', 'slope', 'power'),
            self.dataset.as_dfile(),
        ])

    def get_ys(self, xs):
        background = self._get_param('Background')
        slope = self._get_param('Slope')
        power = self._get_param('Power')
        ys = background + (1.0 - background) * \
            (1 - np.exp(-1.0 * slope * np.power(xs, power)))
        return ys


class Weibull_216(Weibull_215):
    bmds_version_dir = 'BMDS240'
    version = 2.16
    date = '02/28/2013'
    defaults = deepcopy(Weibull_215.defaults)
    defaults['max_iterations']['d'] = 500


class Weibull_217(Weibull_216):
    bmds_version_dir = 'BMDS270'
    version = 2.17
    date = '06/23/2017'


# LOGPROBIT
class LogProbit_32(Dichotomous):
    minimum_dose_groups = 3
    model_name = 'LogProbit'
    bmds_version_dir = 'BMDS231'
    exe = 'probit'
    exe_plot = '10probit'
    version = 3.2
    date = '10/28/2009'
    defaults = {
        'bmdl_curve_calculation': DefaultParams.bmdl_curve_calculation,
        'append_or_overwrite': DefaultParams.append_or_overwrite,
        'smooth_option': DefaultParams.smooth_option,
        'log_transform': DefaultParams.log_transform(d=1),
        'max_iterations': DefaultParams.max_iterations,
        'relative_fn_conv': DefaultParams.relative_fn_conv,
        'parameter_conv': DefaultParams.parameter_conv,
        'background': DefaultParams.param_generator(),
        'slope': DefaultParams.param_generator(name='Slope'),
        'intercept': DefaultParams.param_generator(name='Intercept'),
        'restrict_slope': DefaultParams.restrict(d=1, n='Restrict slope'),
        'bmd_calculation': DefaultParams.bmd_calculation,
        'dose_drop': DefaultParams.dose_drop,
        'bmr': DefaultParams.dich_bmr,
        'bmr_type': DefaultParams.dich_bmr_type,
        'confidence_level': DefaultParams.confidence_level,
    }

    def as_dfile(self):
        self._set_values()
        return '\n'.join([
            self._dfile_print_header_rows(),
            str(self.dataset.dataset_length),
            self._dfile_print_options(
                'max_iterations', 'relative_fn_conv', 'parameter_conv',
                'bmdl_curve_calculation', 'log_transform', 'restrict_slope',
                'bmd_calculation', 'append_or_overwrite', 'smooth_option'),
            self._dfile_print_options(
                'bmr', 'bmr_type', 'confidence_level'),
            self._dfile_print_parameters(
                'background', 'slope', 'intercept'),
            self.dataset.as_dfile(),
        ])

    def get_ys(self, xs):
        background = self._get_param('background')
        slope = self._get_param('slope')
        intercept = self._get_param('intercept')
        ys = background + (1.0 - background) * norm.cdf(intercept + slope * np.log(xs))
        return ys


class LogProbit_33(LogProbit_32):
    bmds_version_dir = 'BMDS240'
    version = 3.3
    date = '02/28/2013'
    defaults = deepcopy(LogProbit_32.defaults)
    defaults['max_iterations']['d'] = 500


class LogProbit_34(LogProbit_33):
    bmds_version_dir = 'BMDS270'
    version = 3.4
    date = '05/21/2017'


# PROBIT
class Probit_32(Dichotomous):
    minimum_dose_groups = 3
    model_name = 'Probit'
    bmds_version_dir = 'BMDS231'
    exe = 'probit'
    exe_plot = '10probit'
    version = 3.2
    date = '10/28/2009'
    defaults = {
        'bmdl_curve_calculation': DefaultParams.bmdl_curve_calculation,
        'append_or_overwrite': DefaultParams.append_or_overwrite,
        'smooth_option': DefaultParams.smooth_option,
        'log_transform': DefaultParams.log_transform(d=0),
        'restrict_slope': DefaultParams.restrict(d=0, n='Restrict slope'),
        'max_iterations': DefaultParams.max_iterations,
        'relative_fn_conv': DefaultParams.relative_fn_conv,
        'parameter_conv': DefaultParams.parameter_conv,
        'background': DefaultParams.param_generator(),
        'slope': DefaultParams.param_generator(name='Slope'),
        'intercept': DefaultParams.param_generator(name='Intercept'),
        'bmd_calculation': DefaultParams.bmd_calculation,
        'dose_drop': DefaultParams.dose_drop,
        'bmr': DefaultParams.dich_bmr,
        'bmr_type': DefaultParams.dich_bmr_type,
        'confidence_level': DefaultParams.confidence_level,
    }

    def as_dfile(self):
        self._set_values()
        return '\n'.join([
            self._dfile_print_header_rows(),
            str(self.dataset.dataset_length),
            self._dfile_print_options(
                'max_iterations', 'relative_fn_conv', 'parameter_conv',
                'bmdl_curve_calculation', 'log_transform', 'restrict_slope',
                'bmd_calculation', 'append_or_overwrite', 'smooth_option'),
            self._dfile_print_options(
                'bmr', 'bmr_type', 'confidence_level'),
            self._dfile_print_parameters(
                'background', 'slope', 'intercept'),
            self.dataset.as_dfile(),
        ])

    def get_ys(self, xs):
        slope = self._get_param('slope')
        intercept = self._get_param('intercept')
        ys = norm.cdf(intercept + slope * xs)
        return ys


class Probit_33(Probit_32):
    bmds_version_dir = 'BMDS240'
    version = 3.3
    date = '02/28/2013'
    defaults = deepcopy(Probit_32.defaults)
    defaults['max_iterations']['d'] = 500


class Probit_34(Probit_33):
    bmds_version_dir = 'BMDS270'
    version = 3.4
    date = '05/21/2017'


# GAMMA
class Gamma_215(Dichotomous):
    minimum_dose_groups = 3
    model_name = 'Gamma'
    bmds_version_dir = 'BMDS231'
    exe = 'gamma'
    exe_plot = '10gammhit'
    version = 2.15
    date = '10/28/2009'
    defaults = {
        'bmdl_curve_calculation': DefaultParams.bmdl_curve_calculation,
        'append_or_overwrite': DefaultParams.append_or_overwrite,
        'smooth_option': DefaultParams.smooth_option,
        'max_iterations': DefaultParams.max_iterations,
        'relative_fn_conv': DefaultParams.relative_fn_conv,
        'parameter_conv': DefaultParams.parameter_conv,
        'background': DefaultParams.param_generator(),
        'slope': DefaultParams.param_generator(name='Slope'),
        'power': DefaultParams.param_generator(name='Power'),
        'bmd_calculation': DefaultParams.bmd_calculation,
        'restrict_power': DefaultParams.restrict(d=1, n='Restrict power'),
        'dose_drop': DefaultParams.dose_drop,
        'bmr': DefaultParams.dich_bmr,
        'bmr_type': DefaultParams.dich_bmr_type,
        'confidence_level': DefaultParams.confidence_level,
    }

    def as_dfile(self):
        self._set_values()
        return '\n'.join([
            self._dfile_print_header_rows(),
            str(self.dataset.dataset_length),
            self._dfile_print_options(
                'max_iterations', 'relative_fn_conv', 'parameter_conv',
                'bmdl_curve_calculation', 'restrict_power',
                'bmd_calculation', 'append_or_overwrite', 'smooth_option'),
            self._dfile_print_options(
                'bmr', 'bmr_type', 'confidence_level'),
            self._dfile_print_parameters(
                'background', 'slope', 'power'),
            self.dataset.as_dfile(),
        ])

    def get_ys(self, xs):
        background = self._get_param('Background')
        slope = self._get_param('Slope')
        power = self._get_param('Power')
        ys = background + (1.0 - background) * gamma.cdf(xs * slope, power)
        return ys


class Gamma_216(Gamma_215):
    bmds_version_dir = 'BMDS240'
    version = 2.16
    date = '02/28/2013'
    defaults = deepcopy(Gamma_215.defaults)
    defaults['max_iterations']['d'] = 500


class Gamma_217(Gamma_216):
    bmds_version_dir = 'BMDS270'
    version = 2.17
    date = '06/22/2017'


# LOGLOGISTIC
class LogLogistic_213(Dichotomous):
    minimum_dose_groups = 3
    model_name = 'LogLogistic'
    bmds_version_dir = 'BMDS231'
    exe = 'logist'
    exe_plot = '10logist'
    version = 2.13
    date = '10/28/2009'
    defaults = {
        'bmdl_curve_calculation': DefaultParams.bmdl_curve_calculation,
        'log_transform': DefaultParams.log_transform(d=1),
        'restrict_slope': DefaultParams.restrict(d=1),
        'append_or_overwrite': DefaultParams.append_or_overwrite,
        'smooth_option': DefaultParams.smooth_option,
        'max_iterations': DefaultParams.max_iterations,
        'relative_fn_conv': DefaultParams.relative_fn_conv,
        'parameter_conv': DefaultParams.parameter_conv,
        'background': DefaultParams.param_generator(),
        'slope': DefaultParams.param_generator(name='Slope'),
        'intercept': DefaultParams.param_generator(name='Intercept'),
        'bmd_calculation': DefaultParams.bmd_calculation,
        'dose_drop': DefaultParams.dose_drop,
        'bmr': DefaultParams.dich_bmr,
        'bmr_type': DefaultParams.dich_bmr_type,
        'confidence_level': DefaultParams.confidence_level,
    }

    def as_dfile(self):
        self._set_values()
        return '\n'.join([
            self._dfile_print_header_rows(),
            str(self.dataset.dataset_length),
            self._dfile_print_options(
                'max_iterations', 'relative_fn_conv', 'parameter_conv',
                'bmdl_curve_calculation', 'log_transform', 'restrict_slope',
                'bmd_calculation', 'append_or_overwrite', 'smooth_option'),
            self._dfile_print_options(
                'bmr', 'bmr_type', 'confidence_level'),
            self._dfile_print_parameters(
                'background', 'slope', 'intercept'),
            self.dataset.as_dfile(),
        ])

    def get_ys(self, xs):
        background = self._get_param('background')
        intercept = self._get_param('intercept')
        slope = self._get_param('slope')
        ys = background + (1.0 - background) / (
            1 + np.exp(-1.0 * intercept - 1.0 * slope * np.log(xs)))
        return ys


class LogLogistic_214(LogLogistic_213):
    bmds_version_dir = 'BMDS240'
    version = 2.14
    date = '02/28/2013'
    defaults = deepcopy(LogLogistic_213.defaults)
    defaults['max_iterations']['d'] = 500


class LogLogistic_215(LogLogistic_214):
    bmds_version_dir = 'BMDS270'
    version = 2.15
    date = '03/20/2017'


# LOGISTIC
class Logistic_213(Dichotomous):
    minimum_dose_groups = 3
    model_name = 'Logistic'
    bmds_version_dir = 'BMDS231'
    exe = 'logist'
    exe_plot = '10logist'
    version = 2.13
    date = '10/28/2009'
    defaults = {
        'bmdl_curve_calculation': DefaultParams.bmdl_curve_calculation,
        'log_transform': DefaultParams.log_transform(d=0),
        'restrict_slope': DefaultParams.restrict(d=0),
        'append_or_overwrite': DefaultParams.append_or_overwrite,
        'smooth_option': DefaultParams.smooth_option,
        'max_iterations': DefaultParams.max_iterations,
        'relative_fn_conv': DefaultParams.relative_fn_conv,
        'parameter_conv': DefaultParams.parameter_conv,
        'background': DefaultParams.param_generator(),
        'slope': DefaultParams.param_generator(name='Slope'),
        'intercept': DefaultParams.param_generator(name='Intercept'),
        'bmd_calculation': DefaultParams.bmd_calculation,
        'dose_drop': DefaultParams.dose_drop,
        'bmr': DefaultParams.dich_bmr,
        'bmr_type': DefaultParams.dich_bmr_type,
        'confidence_level': DefaultParams.confidence_level,
    }

    def as_dfile(self):
        self._set_values()
        return '\n'.join([
            self._dfile_print_header_rows(),
            str(self.dataset.dataset_length),
            self._dfile_print_options(
                'max_iterations', 'relative_fn_conv', 'parameter_conv',
                'bmdl_curve_calculation', 'log_transform', 'restrict_slope',
                'bmd_calculation', 'append_or_overwrite', 'smooth_option'),
            self._dfile_print_options(
                'bmr', 'bmr_type', 'confidence_level'),
            self._dfile_print_parameters(
                'background', 'slope', 'intercept'),
            self.dataset.as_dfile(),
        ])

    def get_ys(self, xs):
        intercept = self._get_param('intercept')
        slope = self._get_param('slope')
        ys = 1.0 / (1.0 + np.exp(-1.0 * intercept - slope * xs))
        return ys


class Logistic_214(Logistic_213):
    bmds_version_dir = 'BMDS240'
    version = 2.14
    date = '02/28/2013'
    defaults = deepcopy(Logistic_213.defaults)
    defaults['max_iterations']['d'] = 500


class Logistic_215(Logistic_214):
    bmds_version_dir = 'BMDS270'
    version = 2.15
    date = '03/20/2017'


# DICHOTOMOUS HILL
class DichotomousHill_13(Dichotomous):
    minimum_dose_groups = 4
    model_name = 'Dichotomous-Hill'
    bmds_version_dir = 'BMDS260'
    exe = 'DichoHill'
    exe_plot = '10DichoHill'
    version = 1.3
    date = '02/28/2013'
    defaults = deepcopy({
        'bmdl_curve_calculation': DefaultParams.bmdl_curve_calculation,
        'restrict_power': DefaultParams.log_transform(d=1),
        'append_or_overwrite': DefaultParams.append_or_overwrite,
        'smooth_option': DefaultParams.smooth_option,
        'max_iterations': DefaultParams.max_iterations,
        'relative_fn_conv': DefaultParams.relative_fn_conv,
        'parameter_conv': DefaultParams.parameter_conv,
        'v': DefaultParams.param_generator(),
        'g': DefaultParams.param_generator(),
        'intercept': DefaultParams.param_generator(),
        'slope': DefaultParams.param_generator(),
        'bmd_calculation': DefaultParams.bmd_calculation,
        'dose_drop': DefaultParams.dose_drop,
        'bmr': DefaultParams.dich_bmr,
        'bmr_type': DefaultParams.dich_bmr_type,
        'confidence_level': DefaultParams.confidence_level,
    })
    defaults['max_iterations']['d'] = 500

    def as_dfile(self):
        self._set_values()
        return '\n'.join([
            self._dfile_print_header_rows(),
            str(self.dataset.dataset_length),
            self._dfile_print_options(
                'max_iterations', 'relative_fn_conv', 'parameter_conv',
                'bmdl_curve_calculation', 'restrict_power',
                'bmd_calculation', 'append_or_overwrite', 'smooth_option'),
            self._dfile_print_options(
                'bmr', 'bmr_type', 'confidence_level'),
            self._dfile_print_parameters(
                'v', 'g', 'intercept', 'slope'),
            self.dataset.as_dfile(),
        ])

    def get_ys(self, xs):
        v = self._get_param('v')
        g = self._get_param('g')
        intercept = self._get_param('intercept')
        slope = self._get_param('slope')
        ys = v * g + (v - v * g) / (1.0 + np.exp(-intercept - slope * np.log(xs)))
        return ys



# -*- coding: utf-8 -*-

# dataset types
DICHOTOMOUS = 'D'
DICHOTOMOUS_CANCER = 'DC'
CONTINUOUS = 'C'
CONTINUOUS_INDIVIDUAL = 'CI'
DTYPES = (
    DICHOTOMOUS,
    DICHOTOMOUS_CANCER,
    CONTINUOUS,
    CONTINUOUS_INDIVIDUAL,
)
DICHOTOMOUS_DTYPES = (
    DICHOTOMOUS,
    DICHOTOMOUS_CANCER,
)
CONTINUOUS_DTYPES = (
    CONTINUOUS,
    CONTINUOUS_INDIVIDUAL,
)

# bmds versions
BMDS231 = 'BMDS231'
BMDS240 = 'BMDS240'
BMDS260 = 'BMDS260'
BMDS2601 = 'BMDS2601'
BMDS270 = 'BMDS270'

# model names
M_Weibull = 'Weibull'
M_LogProbit = 'LogProbit'
M_Probit = 'Probit'
M_Multistage = 'Multistage'
M_Gamma = 'Gamma'
M_Logistic = 'Logistic'
M_LogLogistic = 'LogLogistic'
M_DichotomousHill = 'Dichotomous-Hill'
M_MultistageCancer = 'Multistage-Cancer'
M_Linear = 'Linear'
M_Polynomial = 'Polynomial'
M_Power = 'Power'
M_ExponentialM2 = 'Exponential-M2'
M_ExponentialM3 = 'Exponential-M3'
M_ExponentialM4 = 'Exponential-M4'
M_ExponentialM5 = 'Exponential-M5'
M_Hill = 'Hill'

VARIABLE_POLYNOMIAL = (
    M_Multistage,
    M_MultistageCancer,
    M_Polynomial,
)

# BMR types
DICHOTOMOUS_BMRS = [
    {'type': 'Extra', 'value': 0.1, 'confidence_level': 0.95},
    {'type': 'Added', 'value': 0.1, 'confidence_level': 0.95},
]
CONTINUOUS_BMRS = [
    {'type': 'Std. Dev.', 'value': 1.0, 'confidence_level': 0.95},
    {'type': 'Abs. Dev.', 'value': 0.1, 'confidence_level': 0.95},
    {'type': 'Rel. Dev.', 'value': 0.1, 'confidence_level': 0.95},
    {'type': 'Point', 'value': 1.0, 'confidence_level': 0.95},
    {'type': 'Extra', 'value': 1.0, 'confidence_level': 0.95},
]
BMR_CROSSWALK = {
    DICHOTOMOUS: {
        'Extra': 0,
        'Added': 1
    },
    DICHOTOMOUS_CANCER: {
        'Extra': 0
    },
    CONTINUOUS: {
        'Abs. Dev.': 0,
        'Std. Dev.': 1,
        'Rel. Dev.': 2,
        'Point': 3,
        'Extra': 4
    }
}
BMR_CROSSWALK[CONTINUOUS_INDIVIDUAL] = BMR_CROSSWALK[CONTINUOUS]

# go from integer to human-readable
BMR_INVERTED_CROSSALK = {
    DICHOTOMOUS: {v: k for k, v in BMR_CROSSWALK[DICHOTOMOUS].items()},
    DICHOTOMOUS_CANCER: {v: k for k, v in BMR_CROSSWALK[DICHOTOMOUS_CANCER].items()},
    CONTINUOUS: {v: k for k, v in BMR_CROSSWALK[CONTINUOUS].items()},
}
BMR_INVERTED_CROSSALK[CONTINUOUS_INDIVIDUAL] = BMR_INVERTED_CROSSALK[CONTINUOUS]

# field types
FT_INTEGER = 'i'
FT_DECIMAL = 'd'
FT_BOOL = 'b'
FT_DROPDOSE = 'dd'
FT_RESTRICTPOLY = 'rp'
FT_PARAM = 'p'

# field category
FC_OPTIMIZER = 'op'
FC_OTHER = 'ot'
FC_PARAM = 'p'
FC_BMR = 'b'

# param types
P_DEFAULT = 'd'
P_SPECIFIED = 's'
P_INITIALIZED = 'i'

# logic bins
BIN_NO_CHANGE = 0
BIN_WARNING = 1
BIN_FAILURE = 2
BIN_TYPES = (
    BIN_NO_CHANGE,
    BIN_WARNING,
    BIN_FAILURE,
)
BIN_TEXT = {
    BIN_NO_CHANGE: 'valid',
    BIN_WARNING: 'warning',
    BIN_FAILURE: 'failure',
}
BINMOJI = {
    0: u'✓',
    1: u'?',
    2: u'✕',
}

NULL = '-'



import ctypes
import logging
import os
import sys
import tempfile


logger = logging.getLogger(__name__)


# http://stackoverflow.com/questions/24130623/
# Don't display the Windows GPF dialog if the invoked program dies.
# Required for Weibull model with some datasets with negative slope
SUBPROCESS_FLAGS = 0
if sys.platform.startswith("win"):
    SEM_NOGPFAULTERRORBOX = 0x0002  # From MSDN
    ctypes.windll.kernel32.SetErrorMode(SEM_NOGPFAULTERRORBOX)


class classproperty(property):
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()


class TempFileList(list):
    # Maintains a list of temporary files and cleans up after itself

    def get_tempfile(self, prefix='', suffix='.txt'):
        fd, fn = tempfile.mkstemp(prefix=prefix, suffix=suffix)
        os.close(fd)
        self.append(fn)
        return fn

    def cleanup(self):
        for fn in iter(self):
            try:
                os.remove(fn)
            except OSError:
                pass

    def __del__(self):
        self.cleanup()



import asyncio
from copy import deepcopy
from collections import OrderedDict
import os
import pandas as pd
from simple_settings import settings
import sys

from . import __version__, constants, logic, models, utils, reporter


__all__ = ('BMDS', )


class BMDS(object):
    """
    A single dataset, related models, outputs, and model recommendations.
    """

    @utils.classproperty
    def versions(cls):
        """
        Return all available BMDS software versions.

        Example
        -------

        import bmds

        Returns
        -------
        OrderedDict of available BMDS versions.
        """
        return _BMDS_VERSIONS

    @classmethod
    def get_model(cls, version, model_name):
        """
        Return BMDS model class given BMDS version and model-name.
        """
        models = cls.versions[version].model_options
        for keystore in models.values():
            if model_name in keystore:
                return keystore[model_name]
        raise ValueError('Unknown model name')

    @classmethod
    def get_versions(cls):
        return cls.versions.keys()

    @classmethod
    def latest_version(cls, *args, **kwargs):
        """
        Return the class of the latest version of the BMDS. If additional
        arguments are provided, an instance of this class is generated.
        """
        cls = list(cls.versions.values())[-1]
        if len(args) > 0 or len(kwargs) > 0:
            return cls(*args, **kwargs)
        return cls

    bmr_options = {
        constants.DICHOTOMOUS: constants.DICHOTOMOUS_BMRS,
        constants.DICHOTOMOUS_CANCER: constants.DICHOTOMOUS_BMRS,
        constants.CONTINUOUS: constants.CONTINUOUS_BMRS,
        constants.CONTINUOUS_INDIVIDUAL: constants.CONTINUOUS_BMRS
    }

    def __init__(self, dtype, dataset=None):
        self.dtype = dtype
        if self.dtype not in constants.DTYPES:
            raise ValueError('Invalid data type')
        self.models = []
        self.dataset = dataset

    @property
    def model_options(self):
        raise NotImplementedError('Abstract method requires implementation')

    def get_bmr_options(self):
        return self.bmr_options[self.dtype]

    def get_model_options(self):
        return [
            model.get_default()
            for model in self.model_options[self.dtype].values()
        ]

    @property
    def has_models(self):
        return len(self.models) > 0

    def add_default_models(self, global_overrides=None):

        max_poly_order = min(
            self.dataset.num_dose_groups,
            settings.MAXIMUM_POLYNOMIAL_ORDER + 1)

        for name in self.model_options[self.dtype].keys():
            overrides = deepcopy(global_overrides) \
                if global_overrides is not None \
                else None

            if name in constants.VARIABLE_POLYNOMIAL:
                min_poly_order = 2 if name == constants.M_Polynomial else 1
                for i in range(min_poly_order, max_poly_order):
                    overrides = {} \
                        if overrides is None \
                        else deepcopy(overrides)
                    overrides['degree_poly'] = i
                    self.add_model(name, overrides=overrides)
            else:
                self.add_model(name, overrides=overrides)

    def add_model(self, name, overrides=None, id=None):
        if self.dataset is None:
            raise ValueError('Add dataset to session before adding models')
        Model = self.model_options[self.dtype][name]
        instance = Model(
            dataset=self.dataset,
            overrides=overrides,
            id=id,
        )
        self.models.append(instance)

    async def execute_models(self):
        tasks = [
            model.execute_job()
            for model in self.models
        ]
        await asyncio.wait(tasks)

    def execute(self):
        if sys.platform == 'win32':
            loop = asyncio.ProactorEventLoop()
            asyncio.set_event_loop(loop)
        else:
            loop = asyncio.get_event_loop()
        loop.run_until_complete(self.execute_models())
        loop.close()

    @property
    def recommendation_enabled(self):
        return hasattr(self, 'recommended_model')

    def add_recommender(self, overrides=None):
        self.recommender = logic.Recommender(self.dtype, overrides)

    def recommend(self):
        if not hasattr(self, 'recommender'):
            self.add_recommender()
        self.recommended_model = self.recommender.recommend(self.dataset, self.models)
        self.recommended_model_index = self.models.index(self.recommended_model) \
            if self.recommended_model is not None \
            else None
        return self.recommended_model

    def execute_and_recommend(self, drop_doses=False):
        """
        Execute and recommend a best-fitting model. If drop_doses and no model
        is recommended, drop the highest dose-group and repeat until either:

        1. a model is recommended, or
        2. the dataset is exhausted (i.e., only 3 dose-groups remain).

        This method adds two new attributes to Session:

        1. original_dataset - the unchanged original dataset. The dataset used
            on the object is mutated if doses were dropped
        2. doses_dropped - the number of doses that were dropped

        """
        original_dataset = deepcopy(self.dataset)
        doses_dropped = 0

        self.execute()
        self.recommend()

        if not drop_doses:
            return

        while self.recommended_model is None and \
                self.dataset.num_dose_groups > 3:
            doses_dropped += 1
            self.dataset.drop_dose()
            self.execute()
            self.recommend()

        self.original_dataset = original_dataset
        self.doses_dropped = doses_dropped

    @staticmethod
    def _df_ordered_dict(include_io=True):
        # return an ordered defaultdict list
        keys = [
            'dataset_index', 'model_name', 'model_index',
            'model_version', 'has_output', 'execution_halted',

            'BMD', 'BMDL', 'BMDU', 'CSF',
            'AIC', 'pvalue1', 'pvalue2', 'pvalue3', 'pvalue4',
            'Chi2', 'df', 'residual_of_interest',
            'warnings',

            'logic_bin', 'logic_cautions', 'logic_warnings', 'logic_failures',
            'recommended', 'recommended_variable',
        ]

        if include_io:
            keys.extend(['dfile', 'outfile', 'stdout', 'stderr'])

        return OrderedDict([(key, list()) for key in keys])

    def _add_to_to_ordered_dict(self, d, dataset_index, recommended_only=False):

        for model_index, model in enumerate(self.models):

            # determine if model should be presented, or if a null-model should
            # be presented (if no model is recommended.)
            show_null = False
            if recommended_only:
                if self.recommendation_enabled:
                    if self.recommended_model is None:
                        if model_index == 0:
                            show_null = True
                        else:
                            continue
                    elif self.recommended_model == model:
                        pass
                    else:
                        continue
                else:
                    if model_index == 0:
                        show_null = True
                    else:
                        continue

            d['dataset_index'].append(dataset_index)
            model._to_df(d, model_index, show_null)

    def to_dict(self, dataset_index):
        return dict(
            bmds_version=self.version,
            bmds_python_version=__version__,
            dataset_index=dataset_index,
            dataset=self.dataset.to_dict(),
            models=[model.to_dict(i) for i, model in enumerate(self.models)],
            recommended_model_index=getattr(self, 'recommended_model_index', None)
        )

    def to_excel(self, filename):
        d = self._df_ordered_dict()
        self._add_to_to_ordered_dict(d, 0)
        df = pd.DataFrame(d)
        filename = os.path.expanduser(filename)
        df.to_excel(filename, index=False)

    def save_plots(self, directory, prefix=None,
                   format='png', dpi=None, recommended_only=False):

        directory = os.path.expanduser(directory)
        if not os.path.exists(directory):
            raise ValueError('Directory not found: {}'.format(directory))

        for model in self.models:
            if recommended_only and (self.recommendation_enabled is False or
                                     model.recommended is False):
                continue

            fn = '{}.{}'.format(model.name, format)
            if prefix is not None:
                fn = '{}-{}'.format(prefix, fn)

            fig = model.plot()
            fig.savefig(os.path.join(directory, fn), dpi=dpi)
            fig.clear()

    def to_docx(self, filename=None, title=None,
                input_dataset=True, summary_table=True,
                recommendation_details=True, recommended_model=True,
                all_models=False):
        """
        Write session outputs to a Word file.

        Parameters
        ----------
        filename : str or None
            If provided, the file is saved to this location, otherwise this
            method returns a docx.Document
        input_dataset : bool
            Include input dataset data table
        summary_table : bool
            Include model summary table
        recommendation_details : bool
            Include model recommendation details table
        recommended_model : bool
            Include the recommended model output and dose-response plot, if
            one exists
        all_models : bool
            Include all models output and dose-response plots

        Returns
        -------
        bmds.Reporter
            The bmds.Reporter object.

        """
        rep = reporter.Reporter()
        rep.add_session(self, input_dataset, summary_table,
                        recommendation_details, recommended_model, all_models)

        if filename:
            rep.save(filename)

        return rep

    def _group_models(self):
        """
        If AIC and BMD are numeric and identical, then treat models as
        identical. Returns a list of lists. The outer list is a list of related
        models, the inner list contains each individual model, sorted by the
        number of parameters in ascending order.

        This is required because in some cases, a higher-order model may not use
        some parameters and can effectively collapse to a lower order model
        (for example, a 2nd order polynomial and power model may collapse to a
        linear model). In summary outputs, we may want to present all models in
        one row, since they are the same model effectively.
        """
        od = OrderedDict()

        # Add models to appropriate list. We only aggregate models which
        # completed successfully and have a valid AIC and BMD.
        for i, model in enumerate(self.models):
            output = getattr(model, 'output', {})
            if output.get('AIC') and output.get('BMD') and output['BMD'] > 0:
                key = '{}-{}'.format(output['AIC'], output['BMD'])
                if key in od:
                    od[key].append(model)
                else:
                    od[key] = [model]
            else:
                od[i] = [model]

        # Sort each list by the number of parameters
        def _get_num_params(model):
            return len(model.output['parameters']) \
                if hasattr(model, 'output') and 'parameters' in model.output \
                else 0

        for key, _models in od.items():
            _models.sort(key=_get_num_params)

        return list(od.values())


class BMDS_v231(BMDS):
    version = constants.BMDS231
    version_pretty = 'BMDS v2.3.1'


class BMDS_v240(BMDS_v231):
    version = constants.BMDS240
    version_pretty = 'BMDS v2.4.0'
    model_options = {
        constants.DICHOTOMOUS: OrderedDict([
            (constants.M_Logistic, models.Logistic_214),
            (constants.M_LogLogistic, models.LogLogistic_214),
            (constants.M_Probit, models.Probit_33),
            (constants.M_LogProbit, models.LogProbit_33),
            (constants.M_Multistage, models.Multistage_33),
            (constants.M_Gamma, models.Gamma_216),
            (constants.M_Weibull, models.Weibull_216),
        ]),
        constants.DICHOTOMOUS_CANCER: OrderedDict([
            (constants.M_MultistageCancer, models.MultistageCancer_110),
        ]),
        constants.CONTINUOUS: OrderedDict([
            (constants.M_Linear, models.Linear_217),
            (constants.M_Polynomial, models.Polynomial_217),
            (constants.M_Power, models.Power_217),
            (constants.M_Hill, models.Hill_217),
            (constants.M_ExponentialM2, models.Exponential_M2_19),
            (constants.M_ExponentialM3, models.Exponential_M3_19),
            (constants.M_ExponentialM4, models.Exponential_M4_19),
            (constants.M_ExponentialM5, models.Exponential_M5_19),
        ]),
        constants.CONTINUOUS_INDIVIDUAL: OrderedDict([
            (constants.M_Linear, models.Linear_217),
            (constants.M_Polynomial, models.Polynomial_217),
            (constants.M_Power, models.Power_217),
            (constants.M_Hill, models.Hill_217),
            (constants.M_ExponentialM2, models.Exponential_M2_19),
            (constants.M_ExponentialM3, models.Exponential_M3_19),
            (constants.M_ExponentialM4, models.Exponential_M4_19),
            (constants.M_ExponentialM5, models.Exponential_M5_19),
        ]),
    }


class BMDS_v260(BMDS_v240):
    version = constants.BMDS260
    version_pretty = 'BMDS v2.6.0'
    model_options = {
        constants.DICHOTOMOUS: OrderedDict([
            (constants.M_Logistic, models.Logistic_214),
            (constants.M_LogLogistic, models.LogLogistic_214),
            (constants.M_Probit, models.Probit_33),
            (constants.M_LogProbit, models.LogProbit_33),
            (constants.M_Multistage, models.Multistage_34),
            (constants.M_Gamma, models.Gamma_216),
            (constants.M_Weibull, models.Weibull_216),
            (constants.M_DichotomousHill, models.DichotomousHill_13),
        ]),
        constants.DICHOTOMOUS_CANCER: OrderedDict([
            (constants.M_MultistageCancer, models.MultistageCancer_110),
        ]),
        constants.CONTINUOUS: OrderedDict([
            (constants.M_Linear, models.Linear_220),
            (constants.M_Polynomial, models.Polynomial_220),
            (constants.M_Power, models.Power_218),
            (constants.M_Hill, models.Hill_217),
            (constants.M_ExponentialM2, models.Exponential_M2_110),
            (constants.M_ExponentialM3, models.Exponential_M3_110),
            (constants.M_ExponentialM4, models.Exponential_M4_110),
            (constants.M_ExponentialM5, models.Exponential_M5_110),
        ]),
        constants.CONTINUOUS_INDIVIDUAL: OrderedDict([
            (constants.M_Linear, models.Linear_220),
            (constants.M_Polynomial, models.Polynomial_220),
            (constants.M_Power, models.Power_218),
            (constants.M_Hill, models.Hill_217),
            (constants.M_ExponentialM2, models.Exponential_M2_110),
            (constants.M_ExponentialM3, models.Exponential_M3_110),
            (constants.M_ExponentialM4, models.Exponential_M4_110),
            (constants.M_ExponentialM5, models.Exponential_M5_110),
        ]),
    }


class BMDS_v2601(BMDS_v260):
    version = constants.BMDS2601
    version_pretty = 'BMDS v2.6.0.1'


class BMDS_v270(BMDS_v2601):
    version = constants.BMDS270
    version_pretty = 'BMDS v2.7.0'
    model_options = {
        constants.DICHOTOMOUS: OrderedDict([
            (constants.M_Logistic, models.Logistic_215),
            (constants.M_LogLogistic, models.LogLogistic_215),
            (constants.M_Probit, models.Probit_34),
            (constants.M_LogProbit, models.LogProbit_34),
            (constants.M_Multistage, models.Multistage_34),
            (constants.M_Gamma, models.Gamma_217),
            (constants.M_Weibull, models.Weibull_217),
            (constants.M_DichotomousHill, models.DichotomousHill_13),
        ]),
        constants.DICHOTOMOUS_CANCER: OrderedDict([
            (constants.M_MultistageCancer, models.MultistageCancer_34),
        ]),
        constants.CONTINUOUS: OrderedDict([
            (constants.M_Linear, models.Linear_221),
            (constants.M_Polynomial, models.Polynomial_221),
            (constants.M_Power, models.Power_219),
            (constants.M_Hill, models.Hill_218),
            (constants.M_ExponentialM2, models.Exponential_M2_111),
            (constants.M_ExponentialM3, models.Exponential_M3_111),
            (constants.M_ExponentialM4, models.Exponential_M4_111),
            (constants.M_ExponentialM5, models.Exponential_M5_111),
        ]),
        constants.CONTINUOUS_INDIVIDUAL: OrderedDict([
            (constants.M_Linear, models.Linear_221),
            (constants.M_Polynomial, models.Polynomial_221),
            (constants.M_Power, models.Power_219),
            (constants.M_Hill, models.Hill_218),
            (constants.M_ExponentialM2, models.Exponential_M2_111),
            (constants.M_ExponentialM3, models.Exponential_M3_111),
            (constants.M_ExponentialM4, models.Exponential_M4_111),
            (constants.M_ExponentialM5, models.Exponential_M5_111),
        ]),
    }


_BMDS_VERSIONS = OrderedDict((
    (constants.BMDS231, BMDS_v231),
    (constants.BMDS240, BMDS_v240),
    (constants.BMDS260, BMDS_v260),
    (constants.BMDS2601, BMDS_v2601),
    (constants.BMDS270, BMDS_v270),
))
